import logging
from dataclasses import dataclass

import numpy as np

from .air_to_wall import AirToWallTransfer

logger = logging.getLogger(__name__)


@dataclass
class AirSourceHeatExchangerGeometry:
    """
    Geometry of a fin and tube heat exchanger with two rows of pipes in a shifted arrangement.

    Source: VDI-Wärmeatlas, Berechnungsblätter für den Wärmeübergang, 11. Auflage, S.1461

    As the source is in German, the names are kept in German as well.
    """
    t_l: float  # Achsabstand der Rohre in Luftrichtung in m
    t_q: float    # Achsabstand der Rohre quer zur Luftrichtung in m
    tiefe: float  # Tiefe der Rippe gesamt
    d_a: float  # Äußerer Rohrdurchmesser
    d_i: float  # Innener Rohrdurchmesser
    lambda_R: float  # Wärmeleitfähigkeit Material der Wand
    n_Rohre: int = 50
    n_Rippen: int = 500
    a: float = 1.95e-3
    dicke_rippe: float = 0.05e-3
    laenge: float = 1.040
    hoehe: float = 0.64

    @property
    def char_length(self) -> float:
        """Return the characteristic length d_a."""
        return self.d_a

    @property
    def A_Rippen(self) -> float:
        """Return the total surface area of the fins."""
        return 2 * self.n_Rippen * (self.tiefe * self.hoehe - self.n_Rohre * np.pi * 0.25 * self.d_a ** 2)

    @property
    def A(self) -> float:
        """Total air side heat transfer area."""
        return self.A_Rippen + self.A_FreieOberflaecheRohr

    @property
    def A_FreieOberflaecheRohr(self) -> float:
        """Free air side area of the tubes."""
        return self.n_Rippen * np.pi * self.d_a * self.a * self.n_Rohre

    @property
    def A_RohrInnen(self) -> float:
        """Total refrigerant heat transfer area."""
        return self.laenge * self.d_i * np.pi * self.n_Rohre

    @property
    def A_RohrUnberippt(self) -> float:
        """Total outside area of the tubes without fins"""
        return self.laenge * self.d_a * np.pi * self.n_Rohre

    @property
    def verjuengungsfaktor(self) -> float:
        """Reduction factor A_cross_free/A_cross_smallest"""
        return ((self.hoehe * self.laenge) /
                (self.hoehe * self.laenge -
                 self.hoehe * self.n_Rippen * self.dicke_rippe -
                 self.d_a * self.n_Rohre / 2 * (self.laenge - self.n_Rippen * self.dicke_rippe)))

    @property
    def phi(self) -> float:
        """Auxiliary variable for fin efficiency"""
        if self.t_l >= 0.5 * self.t_q:
            b_r = self.t_q
        else:
            b_r = 2 * self.t_l
        l_r = np.sqrt(self.t_l **2 + self.t_q **2 / 4)
        phi_strich = 1.27 * b_r / self.d_a * np.sqrt(l_r / b_r - 0.3)
        return (phi_strich - 1) * (1 + 0.35 * np.log(phi_strich))

    def eta_R(self, alpha_R) -> float:
        """
        Calculate fin efficiency.

        Args:
            alpha_R (float): Average heat transfer coefficient for tube and fin.

        Returns:
            float: Fin efficiency.
        """
        X = self.phi * self.d_a / 2 * np.sqrt(2 * alpha_R / (self.lambda_R * self.dicke_rippe))
        return 1 / X * (np.exp(X) - np.exp(-X)) / (np.exp(X) + np.exp(-X))

    def alpha_S(self, alpha_R) -> float:
        """
        Calculate apparent heat transfer coefficient.

        Args:
            alpha_R (float): Average heat transfer coefficient for tube and fin.

        Returns:
            float: Apparent heat transfer coefficient.
        """
        A_R_to_A = self.A_Rippen / self.A
        return alpha_R * (1 - (1 - self.eta_R(alpha_R)) * A_R_to_A)


class VDIAtlasAirToWallTransfer(AirToWallTransfer):
    """
    VDI-Atlas based heat transfer estimation.
    Check `AirToWallTransfer` for further argument options.

    This class assumes two pipes with a shifted arrangement.

    Args:
        A_cross (float): Free cross-sectional area.
        characteristic_length (float): Characteristic length d_a.
        geometry_parameters (AirSourceHeatExchangerGeometry):
            Geometry parameters for heat exchanger according to VDI-Atlas
    """

    def __init__(self,
                 A_cross: float, characteristic_length: float,
                 geometry_parameters: AirSourceHeatExchangerGeometry):
        super().__init__(A_cross=A_cross, characteristic_length=characteristic_length)
        self.geometry_parameters = geometry_parameters

    def calc_reynolds(self, dynamic_viscosity: float, m_flow: float) -> float:
        """
        Calculate Reynolds number.

        Args:
            dynamic_viscosity (float): Dynamic viscosity of the fluid.
            m_flow (float): Mass flow rate.

        Returns:
            float: Reynolds number.
        """
        velocity_times_dens = m_flow / self.A_cross * self.geometry_parameters.verjuengungsfaktor
        return (velocity_times_dens * self.characteristic_length) / dynamic_viscosity

    def calc_laminar_area_nusselt(self, Re: float, Pr: float, lambda_: float) -> float:
        """
        Calculate apparent heat transfer coefficient based on Nusselt correlation.

        Args:
            Re (float): Reynolds number.
            Pr (float): Prandtl number.
            lambda_ (float): Thermal conductivity of air.

        Returns:
            float: Apparent heat transfer coefficient.
        """
        C = 0.33  # Versetzte Anordnung, zwei Rohre
        if Re < 1e3 or Re > 1e5:
            logger.debug("Given Re %s is outside of allowed bounds for VDI-Atlas", Re)
        A_div_A_0 = self.geometry_parameters.A / self.geometry_parameters.A_RohrUnberippt
        if A_div_A_0 < 5 or A_div_A_0 > 30:
            logger.debug("Given A/A0 is outside of bounds for method VDI-Atlas")
        Nu = (
                C * Re ** 0.6 *
                A_div_A_0 ** (-0.15) *
                Pr ** (1 / 3)
        )
        alpha_R = Nu * lambda_ / self.characteristic_length
        alpha_S = self.geometry_parameters.alpha_S(alpha_R=alpha_R)
        return alpha_S



import warnings
from abc import ABC
import numpy as np
import pandas as pd

from vclibpy.components.compressors.compressor import Compressor
from vclibpy.datamodels import Inputs


def calc_ten_coefficients(T_eva, T_con, coef_list):
    """
    Calculate the result using the ten-coefficient method.

    Args:
        T_eva (float): Evaporator temperature in Celsius.
        T_con (float): Condenser temperature in Celsius.
        coef_list (list): List of coefficients.

    Returns:
        float: Result of the calculation.
    """
    # Formula for the ten-coefficient method according to the datasheet
    z = coef_list[0] + coef_list[1] * T_eva + coef_list[2] * T_con + coef_list[3] * T_eva ** 2 + \
        coef_list[4] * T_eva * T_con + coef_list[5] * T_con ** 2 + coef_list[6] * T_eva ** 3 + \
        coef_list[7] * T_eva ** 2 * T_con + coef_list[8] * T_con ** 2 * T_eva + coef_list[9] * T_con ** 3
    return z


class BaseTenCoefficientCompressor(Compressor, ABC):
    """
    Base class for compressors using the ten-coefficient method.

    Used table has to be in this format
    (order of the columns is not important).
    The values must be the same as in the example tabel.
    The column names can be different but must
    then be matched with argument parameter_names.
    (All typed in numbers are fictional placeholders)

          Capacity(W)     Input Power(W)      Flow Rate(kg/h)     Capacity(W)     ...     Flow Rate(kg/h)
    n     n1              n1                  n1                  n2              ...     n_last
    P1    42              12                  243                 32              ...     412
    ...   ...             ...                 ...                 ...             ...     ...
    P10   10              23                  21                  41              ...     2434

    Args:
        N_max (float): Maximal rotations per second of the compressor.
        V_h (float): Volume of the compressor in m^3.
        datasheet (str): Path of the datasheet file.
        **kwargs:
            parameter_names (dict, optional):
                Dictionary to match internal parameter names (keys) to the names used in the table values.
                Default
                {
                    "m_flow": "Flow Rate(kg/h)",
                    "capacity": "Capacity(W)",
                    "input_power": "Input Power(W)",
                    "eta_s": "Isentropic Efficiency(-)",
                    "lambda_h": "Volumentric Efficiency(-)",
                    "eta_mech": "Mechanical Efficiency(-)"
                }
            sheet_name (str, optional): Name of the sheet in the datasheet. Defaults to None.
    """

    def __init__(self, N_max, V_h, datasheet, **kwargs):
        """
        Initialize the BaseTenCoefficientCompressor.

        Args:
            N_max (float): Maximal rotations per second of the compressor.
            V_h (float): Volume of the compressor in m^3.
            datasheet (str): Path of the datasheet file.
            parameter_names (dict, optional): Dictionary of parameter names. Defaults to None.
            sheet_name (str, optional): Name of the sheet in the datasheet. Defaults to None.
        """

        super().__init__(N_max, V_h)
        sheet_name = kwargs.get('sheet_name', None)
        self.md = pd.read_excel(datasheet, sheet_name=sheet_name)
        parameter_names = kwargs.get('parameter_names', None)
        if parameter_names is None:
            self.parameter_names = {
                "m_flow": "Flow Rate(kg/h)",
                "capacity": "Capacity(W)",
                "input_power": "Input Power(W)",
                "eta_s": "Isentropic Efficiency(-)",
                "lambda_h": "Volumentric Efficiency(-)",
                "eta_mech": "Mechanical Efficiency(-)"
            }
        else:
            self.parameter_names = parameter_names

    def get_parameter(self, T_eva, T_con, n, type_):
        """
        Get a parameter based on temperatures, rotations, and parameter type from the datasheet.

        Args:
            T_eva (float): Evaporator temperature in Celsius.
            T_con (float): Condenser temperature in Celsius.
            n (float): Rotations per second.
            type_ (str): Parameter type in parameter_names.

        Returns:
            float: Interpolated parameter value.
        """
        param_list = []
        n_list = []

        sampling_points = sum(
            self.parameter_names[type_] in s for s in list(self.md.columns.values))  # counts number of sampling points

        for i in range(sampling_points):
            if i == 0:
                coefficients = self.md[self.parameter_names[type_]].tolist()
            else:
                coefficients = self.md[str(self.parameter_names[type_] + "." + str(i))].tolist()
            n_list.append(coefficients.pop(0))
            param_list.append(calc_ten_coefficients(T_eva, T_con, coefficients))

        return np.interp(self.get_n_absolute(n), n_list, param_list)  # linear interpolation


class TenCoefficientCompressor(BaseTenCoefficientCompressor):
    """
    Compressor based on the ten coefficient method.

    Used table has to be in this format
    (order of the columns is not important).
    The values must be the same as in the example tabel.
    The column names can be different but must
    then be matched with the keyword argument parameter_names.
    (All typed in numbers are fictional placeholders)

          Capacity(W)     Input Power(W)      Flow Rate(kg/h)     Capacity(W)     ...     Flow Rate(kg/h)
    n     n1              n1                  n1                  n2              ...     n_last
    P1    42              12                  243                 32              ...     412
    ...   ...             ...                 ...                 ...             ...     ...
    P10   10              23                  21                  41              ...     2434

    T_sh and T_sc have to be set according to the data sheet of your compressor. capacity_definition defines the
    parameter "capacity" in the datasheet. If capacity is the specific cooling capacity (h1-h4), set it on "cooling".
    If capacity is the specific heating capacity (h2-h3), set it on "heating".
    In the case of cooling capacity, the mechanical efficiency of the compressor has to be assumed (assumed_eta_mech)
    as h2 can't be calculated otherwise. Summary:
        - For the case heating capacity: h2 = h3 + capacity / m_flow
        - For the case cooling capacity:  h2 = h3 + (capacity + p_el * assumed_eta_mech) / m_flow

    Args:
        N_max (float): Maximal rotations per second of the compressor.
        V_h (float): Volume of the compressor in m^3.
        T_sc (float): Subcooling according to datasheet in K.
        T_sh (float): Superheating according to datasheet in K.
        capacity_definition (str): Definition of "capacity" in the datasheet. "cooling" or "heating".
        assumed_eta_mech (float): Assumed mechanical efficiency of the compressor (only needed if cooling).
        datasheet (str): Path of the modified datasheet.
        **kwargs:
            parameter_names (dict, optional):
                Dictionary to match internal parameter names (keys) to the names used in the table values.
                Default
                {
                    "m_flow": "Flow Rate(kg/h)",
                    "capacity": "Capacity(W)",
                    "input_power": "Input Power(W)"
                }
            sheet_name (str, optional): Name of the sheet in the datasheet. Defaults to None.
    """

    def __init__(self, N_max, V_h, T_sc, T_sh, capacity_definition, assumed_eta_mech, datasheet, **kwargs):
        super().__init__(N_max=N_max, V_h=V_h, datasheet=datasheet, **kwargs)
        self.T_sc = T_sc
        self.T_sh = T_sh
        if capacity_definition not in ["cooling", "heating"]:
            raise ValueError("capacity_definition has to be either 'heating' or 'cooling'")
        self._capacity_definition = capacity_definition
        self.assumed_eta_mech = assumed_eta_mech
        self.datasheet = datasheet

    def get_lambda_h(self, inputs: Inputs):
        """
        Get the volumetric efficiency.

        Args:
            inputs (Inputs): Input parameters.

        Returns:
            float: Volumetric efficiency.
        """
        p_outlet = self.get_p_outlet()

        n_abs = self.get_n_absolute(inputs.n)
        T_eva = self.med_prop.calc_state("PQ", self.state_inlet.p, 1).T - 273.15  # [°C]
        T_con = self.med_prop.calc_state("PQ", p_outlet, 0).T - 273.15  # [°C]

        if round((self.state_inlet.T - T_eva - 273.15), 2) != round(self.T_sh, 2):
            warnings.warn("The superheating of the given state is not "
                          "equal to the superheating of the datasheet. "
                          "State1.T_sh= " + str(round((self.state_inlet.T - T_eva - 273.15), 2)) +
                          ". Datasheet.T_sh = " + str(self.T_sh))
        # The datasheet has a given superheating temperature which can
        # vary from the superheating of the real state 1
        # which is given by the user.
        # Thus a new self.state_inlet_datasheet has to
        # be defined for all further calculations
        state_inlet_datasheet = self.med_prop.calc_state("PT", self.state_inlet.p, T_eva + 273.15 + self.T_sh)

        m_flow = self.get_parameter(T_eva, T_con, inputs.n, "m_flow") / 3600  # [kg/s]

        lambda_h = m_flow / (n_abs * state_inlet_datasheet.d * self.V_h)
        return lambda_h

    def get_eta_isentropic(self, p_outlet: float, inputs: Inputs):
        """
        Get the isentropic efficiency.

        Args:
            p_outlet (float): Outlet pressure in Pa.
            inputs (Inputs): Input parameters.

        Returns:
            float: Isentropic efficiency.
        """
        T_con, state_inlet_datasheet, m_flow, capacity, p_el = self._calculate_values(
            p_2=p_outlet, inputs=inputs
        )

        h3 = self.med_prop.calc_state("PT", p_outlet, T_con + 273.15 - self.T_sc).h  # [J/kg]
        h2s = self.med_prop.calc_state("PS", p_outlet, state_inlet_datasheet.s).h  # [J/kg]

        if self._capacity_definition == "heating":
            h2 = h3 + capacity / m_flow  # [J/kg]
        else:
            h2 = h3 + (capacity + p_el * self.assumed_eta_mech) / m_flow  # [J/kg]

        if h2s > h2:
            raise ValueError("The calculated eta_s is above 1. You probably chose the wrong capacity_definition")

        eta_s = (h2s - state_inlet_datasheet.h) / (h2 - state_inlet_datasheet.h)
        return eta_s

    def get_eta_mech(self, inputs: Inputs):
        """
        Get the mechanical efficiency.

        Args:
            inputs (Inputs): Input parameters.

        Returns:
            float: Mechanical efficiency.
        """
        p_outlet = self.get_p_outlet()

        if self._capacity_definition == "cooling":
            return self.assumed_eta_mech
        # Else heating
        T_con, state_inlet_datasheet, m_flow, capacity, p_el = self._calculate_values(
            p_2=p_outlet, inputs=inputs
        )

        h3 = self.med_prop.calc_state("PT", p_outlet, T_con + 273.15 - self.T_sc).h  # [J/kg]
        h2 = h3 + capacity / m_flow  # [J/kg]

        eta_mech = (m_flow * (h2 - state_inlet_datasheet.h)) / p_el
        return eta_mech

    def _calculate_values(self, p_2: float, inputs: Inputs):
        """
        Calculate intermediate values for efficiency calculations.

        Args:
            p_2 (float): Outlet pressure in Pa.
            inputs (Inputs): Input parameters.

        Returns:
            Tuple[float, State, float, float, float]: Intermediate values.
        """
        T_eva = self.med_prop.calc_state("PQ", self.state_inlet.p, 1).T - 273.15  # [°C]
        T_con = self.med_prop.calc_state("PQ", p_2, 0).T - 273.15  # [°C]

        state_inlet_datasheet = self.med_prop.calc_state("PT", self.state_inlet.p, T_eva + 273.15 + self.T_sh)

        m_flow = self.get_parameter(T_eva, T_con, inputs.n, "m_flow") / 3600  # [kg/s]
        capacity = self.get_parameter(T_eva, T_con, inputs.n, "capacity")  # [W]
        p_el = self.get_parameter(T_eva, T_con, inputs.n, "input_power")  # [W]
        return T_con, state_inlet_datasheet, m_flow, capacity, p_el


class DataSheetCompressor(BaseTenCoefficientCompressor):
    """
    Compressor based on the ten coefficient method.

    Used table has to be in this format
    (order of the columns is not important).
    The values must be the same as in the example tabel.
    The column names can be different but must
    then be matched with the keyword argument parameter_names.
    (All typed in numbers are fictional placeholders)

          Isentropic      Volumetric        Mechanical      Isentropic              Mechanical
          Efficiency(-)   Efficiency(-)     Efficiency(-)   Efficiency(-)   ...     Efficiency(-)
    n     n1              n1                n1              n2              ...     n_last
    P1    42              12                243             32              ...     412
    ...   ...             ...               ...             ...             ...     ...
    P10   10              23                21              41              ...     2434

    Args:
        N_max (float): Maximal rotations per second of the compressor.
        V_h (float): Volume of the compressor in m^3.
        datasheet (str): Path of the datasheet file.
        **kwargs:
            parameter_names (dict, optional):
                Dictionary to match internal parameter names (keys) to the names used in the table values.
                Default
                {
                    "eta_s": "Isentropic Efficiency(-)",
                    "lambda_h": "Volumetric Efficiency(-)",
                    "eta_mech": "Mechanical Efficiency(-)"
                }
            sheet_name (str, optional): Name of the sheet in the datasheet. Defaults to None.
    """

    def __init__(self, N_max, V_h, datasheet, **kwargs):
        super().__init__(N_max=N_max, V_h=V_h, datasheet=datasheet, **kwargs)

    def get_lambda_h(self, inputs: Inputs):
        """
        Get the volumetric efficiency.

        Args:
            inputs (Inputs): Input parameters.

        Returns:
            float: Volumetric efficiency.
        """
        p_outlet = self.get_p_outlet()
        T_eva = self.med_prop.calc_state("PQ", self.state_inlet.p, 0).T
        T_con = self.med_prop.calc_state("PQ", p_outlet, 0).T
        return self.get_parameter(T_eva, T_con, inputs.n, "lambda_h")

    def get_eta_isentropic(self, p_outlet: float, inputs: Inputs):
        """
        Get the isentropic efficiency.

        Args:
            p_outlet (float): Outlet pressure in Pa.
            inputs (Inputs): Input parameters.

        Returns:
            float: Isentropic efficiency.
        """
        T_eva = self.med_prop.calc_state("PQ", self.state_inlet.p, 0).T
        T_con = self.med_prop.calc_state("PQ", p_outlet, 0).T
        return self.get_parameter(T_eva, T_con, inputs.n, "eta_s")

    def get_eta_mech(self, inputs: Inputs):
        """
        Get the mechanical efficiency.

        Args:
            inputs (Inputs): Input parameters.

        Returns:
            float: Mechanical efficiency.
        """
        p_outlet = self.get_p_outlet()
        T_eva = self.med_prop.calc_state("PQ", self.state_inlet.p, 0).T
        T_con = self.med_prop.calc_state("PQ", p_outlet, 0).T
        return self.get_parameter(T_eva, T_con, inputs.n, "eta_mech")



from typing import List
import csv
import pandas as pd

from vclibpy.components.compressors import TenCoefficientCompressor
from vclibpy import media, Inputs

try:
    from sklearn.linear_model import LinearRegression
    from sklearn.preprocessing import PolynomialFeatures
    from xlsxwriter.workbook import Workbook
except ImportError as err:
    raise ImportError("You have to install xlsxwriter and "
                      "sklearn to use this regression tool")


def create_regression_data(
        variables: List[str],
        T_con: List[float], T_eva: List[float], n: List[float],
        T_sh: float, T_sc: float, n_max: int, V_h: float, fluid: str, datasheet: str,
        capacity_definition: str, assumed_eta_mech: int,
        folder_path: str):
    """
    Performs multidimensional linear regression to create compressor learning data.

    Args:
        variables: (List[str]):
            Variable names to create regressions for.
            Options are: eta_s, lambda_h, and eta_mech
        T_con (List[float]): Condensing temperatures in K
        T_eva (List[float]): Evaporation temperatures in K
        n (List[float]): Compressor speeds from 0 to 1
        T_sh (float): Superheat temperature.
        T_sc (float): Subcooling temperature.
        n_max (int): Maximum compressor speed.
        V_h (float): Compressor volume.
        fluid (str): Type of refrigerant.
        datasheet (str): Path to the modified datasheet.
        capacity_definition (str): Definition of compressor capacity (e.g., "cooling").
        assumed_eta_mech (int): Assumed mechanical efficiency.
        folder_path (str): Path to the folder where the newly created table will be saved.

    Returns:
        List[float]: A list containing the regression parameters [P0, P1, ..., P9].

    Raises:
        ValueError: If any specified variable column is not present in the DataFrame.

    Example:
        >>> create_regression_data(11.1, 8.3, 120, 20.9e-6, "PROPANE", "path/to/datasheet.xlsx",
        ...                        "cooling", 1, 6, 5, 5, 5, 303.15, 243.15, "path/to/save", 3)
        [intercept, P1, P2, P3, P4, P5, P6, P7, P8, P9]
    """
    # create RefProp, fluid & compressor instance
    med_prop = media.CoolProp(fluid)

    compressor = TenCoefficientCompressor(
        N_max=n_max, V_h=V_h, T_sc=T_sc, T_sh=T_sh,
        capacity_definition=capacity_definition,
        assumed_eta_mech=assumed_eta_mech,
        datasheet=datasheet
    )

    compressor.med_prop = med_prop
    keywords = {
        "eta_s": "Isentropic Efficiency(-)",
        "lambda_h": "Volumentric Efficiency(-)",
        "eta_mech": "Mechanical Efficiency(-)"
    }

    tuples_for_cols = [("", "n")]
    for _variable in variables:
        for _n in n:
            tuples_for_cols.append((keywords[_variable], compressor.get_n_absolute(_n)))
    # tuples_for_cols:
    #     eta_s   eta_s   eta_s   lambda_h    lambda_h    lambda_h    eta_mech    ...
    # n   30      60      90      30          60          90          30          ...
    cols = pd.MultiIndex.from_tuples(tuples_for_cols)
    final_df = pd.DataFrame(
        data={cols[0]: ["P1", "P2", "P3", "P4", "P5", "P6", "P7", "P8", "P9", "P10"]},
        columns=cols
    )
    # final_df: column names are tuples (tuples_for_cols).
    # First column is filled with P1, P2, ...

    # for-loop for multiple types(eta_s, eta_mech, etc)
    for m, _variable in enumerate(variables):
        for k, _n in enumerate(n):  # for-loop for multiple rotation speeds
            T_eva_list = []
            T_con_list = []
            result_list = []
            # for-loop for multiple evaporating temperatures
            for i in range(len(T_eva)):
                # for-loop for multiple condensing temperatures
                for j in range(len(T_con)):
                    if T_eva[i] < T_con[j]:
                        p1 = med_prop.calc_state("TQ", T_eva[i], 1).p
                        state_1 = med_prop.calc_state("PT", p1, (T_eva[i] + T_sh))
                        compressor.state_inlet = state_1
                        p2 = med_prop.calc_state("TQ", T_con[j], 1).p
                        # The enthalpy and entropy of the outlet
                        # state do not matter, only the pressure:
                        # TODO: Enable calculation of get_lambda_h etc. with p2 only
                        state_2 = med_prop.calc_state("PS", p2, state_1.s)
                        compressor.state_outlet = state_2
                        T_eva_list.append(T_eva[i])
                        T_con_list.append(T_con[j])
                        inputs = Inputs(n=_n)

                        if _variable == "eta_s":
                            result_list.append(compressor.get_eta_isentropic(
                                p_outlet=p2, inputs=inputs)
                            )
                        elif _variable == "lambda_h":
                            result_list.append(compressor.get_lambda_h(inputs=inputs))
                        elif _variable == "eta_mech":
                            result_list.append(compressor.get_eta_mech(inputs=inputs))

            df = pd.DataFrame(
                data={"T_eva": T_eva_list,
                      "T_con": T_con_list,
                      _variable: result_list}
            )

            final_df[cols[m * len(n) + k + 1]] = create_regression_parameters(df, _variable)

    # dataframes with a double column header can't be saved as a
    # .xlsx yet, if index = False (NotImplementedError).
    # .xlsx format is necessary, because TenCoefficientCompressor.get_parameter()
    # expects a .xlsx as an input
    # --> workaround: save the dataframe as a .csv, read it again and save it as a .xlsx
    # TODO: Revert once this feature is in pandas.
    final_df.to_csv(folder_path + r"\regressions.csv", index=False)

    workbook = Workbook(folder_path + r"\regressions.xlsx")
    worksheet = workbook.add_worksheet()
    with open(folder_path + r"\regressions.csv", 'rt', encoding='utf8') as f:
        reader = csv.reader(f)
        for r, row in enumerate(reader):
            for c, col in enumerate(row):
                worksheet.write(r, c, col)
    workbook.close()


def create_regression_parameters(df: pd.DataFrame, variable: str):
    """
    Performs multidimensional linear regression to calculate
    ten coefficient regression parameters.

    Args:
        df (pd.DataFrame): The input DataFrame containing the necessary columns.
        variable (str): The column name for the dependent variable.

    Returns:
        List[float]: A list containing the ten regression parameters.

    Raises:
        ValueError: If the specified variable column is not present in the DataFrame.

    Example:
        >>> df = pd.DataFrame({'T_eva': [1, 2, 3], 'T_con': [4, 5, 6], 'X': [7, 8, 9]})
        >>> create_regression_parameters(df, 'X')
        [intercept, P1, P2, P3, P4, P5, P6, P7, P8, P9]
    """
    # extract the columns x, y und z
    x = df['T_eva'].values
    y = df['T_con'].values
    z = df[variable].values

    # define the features (x, y, x^2, xy, y^2, x^3, x^2y, xy^2, y^3)
    features = PolynomialFeatures(degree=3, include_bias=False)
    X = features.fit_transform(pd.concat([pd.DataFrame(x), pd.DataFrame(y)], axis=1))

    # Execute the multidimensional linear regression
    model = LinearRegression().fit(X, z)

    output = [model.intercept_, model.coef_[0], model.coef_[1], model.coef_[2],
              model.coef_[3], model.coef_[4],
              model.coef_[5], model.coef_[6], model.coef_[7], model.coef_[8]]
    # output = P1-P10
    return output



import unittest
from random import random
import vclibpy


class TestDataModels(unittest.TestCase):
    """Test-class for the datamodels module of vclibpy."""

    def setUp(self) -> None:
        self.fs_state = vclibpy.FlowsheetState()
        self.var = vclibpy.datamodels.Variable(name="test", value=random(), unit="K", description="dummy")

    def test_variable_set(self):
        self.fs_state.set(**self.var.__dict__)
        self.assertEqual(self.fs_state.test, self.var.value)
        self.assertEqual(self.fs_state.get_variables()["test"], self.var)
        self.assertEqual(self.fs_state.get_variable_names(), ["test"])
        self.assertEqual(self.fs_state._variables, {"test": self.var})
        # Update value
        self.fs_state.set(name="test", value=0)
        self.assertEqual(self.fs_state.test, 0)


if __name__ == "__main__":
    unittest.main()



from .heat_transfer import TwoPhaseHeatTransfer, HeatTransfer, calc_reynolds_pipe
from vclibpy.components.heat_exchangers.heat_transfer import constant
from vclibpy.components.heat_exchangers.heat_transfer import wall



import abc

from typing import Tuple

from vclibpy import media
from vclibpy.datamodels import FlowsheetState, Inputs
from vclibpy.components.component import BaseComponent
from vclibpy.components.heat_exchangers.heat_transfer.heat_transfer import HeatTransfer, TwoPhaseHeatTransfer


class HeatExchanger(BaseComponent, abc.ABC):
    """
    Class for a heat exchanger.

    Args:
        A (float):
            Area of HE in m^2 for NTU calculation
        secondary_medium (str):
            Name for secondary medium, e.g. `water` or `air`
        wall_heat_transfer (HeatTransfer):
            Model for heat transfer inside wall
        secondary_heat_transfer (HeatTransfer):
            Model for heat transfer from secondary medium to wall
        gas_heat_transfer (HeatTransfer):
            Model for heat transfer from refrigerant gas to wall
        liquid_heat_transfer (HeatTransfer):
            Model for heat transfer from refrigerant liquid to wall
        two_phase_heat_transfer (TwoPhaseHeatTransfer):
            Model for heat transfer from refrigerant two phase to wall
    """

    def __init__(
            self,
            A: float,
            wall_heat_transfer: HeatTransfer,
            secondary_heat_transfer: HeatTransfer,
            gas_heat_transfer: HeatTransfer,
            liquid_heat_transfer: HeatTransfer,
            two_phase_heat_transfer: TwoPhaseHeatTransfer,
            secondary_medium: str
    ):
        super().__init__()
        self.A = A
        self.secondary_medium = secondary_medium.lower()

        self._wall_heat_transfer = wall_heat_transfer
        self._secondary_heat_transfer = secondary_heat_transfer
        self._gas_heat_transfer = gas_heat_transfer
        self._liquid_heat_transfer = liquid_heat_transfer
        self._two_phase_heat_transfer = two_phase_heat_transfer

        self.med_prop_sec = None  # Later start in start_secondary_med_prop
        self._m_flow_secondary = None
        self._secondary_cp = 0  # Allow initial calculation of _m_flow_secondary_cp if cp is not set
        self._m_flow_secondary_cp = 0

    def start_secondary_med_prop(self):
        """
        Set up the wrapper for the secondary medium's media properties.
        """
        # Set up the secondary_medium wrapper:
        med_prop_class, med_prop_kwargs = media.get_global_med_prop_and_kwargs()
        if self.secondary_medium == "air" and med_prop_class == media.RefProp:
            fluid_name = "AIR.PPF"
        else:
            fluid_name = self.secondary_medium
        if self.med_prop_sec is not None:
            if self.med_prop_sec.fluid_name == fluid_name:
                return
            self.med_prop_sec.terminate()
        self.med_prop_sec = med_prop_class(fluid_name=self.secondary_medium, **med_prop_kwargs)

    def terminate_secondary_med_prop(self):
        if self.med_prop_sec is not None:
            self.med_prop_sec.terminate()

    @abc.abstractmethod
    def calc(self, inputs: Inputs, fs_state: FlowsheetState) -> Tuple[float, float]:
        """
        Calculate the heat exchanger based on the given inputs.

        The flowsheet state can be used to save important variables
        during calculation for later analysis.

        Both return values are used to check if the heat transfer is valid or not.

        Args:
            inputs (Inputs): The inputs for the calculation.
            fs_state (FlowsheetState): The flowsheet state to save important variables.

        Returns:
            Tuple[float, float]:
                error: Error in percentage between the required and calculated heat flow rates.
                dT_min: Minimal temperature difference (can be negative).
        """
        raise NotImplementedError

    def calc_alpha_two_phase(self, state_q0, state_q1, inputs: Inputs, fs_state: FlowsheetState) -> float:
        """
        Calculate the two-phase heat transfer coefficient.

        Args:
            state_q0: State at vapor quality 0.
            state_q1: State at vapor quality 1.
            inputs (Inputs): The inputs for the calculation.
            fs_state (FlowsheetState): The flowsheet state to save important variables.

        Returns:
            float: The two-phase heat transfer coefficient.
        """
        return self._two_phase_heat_transfer.calc(
            state_q0=state_q0,
            state_q1=state_q1,
            inputs=inputs,
            fs_state=fs_state,
            m_flow=self.m_flow,
            med_prop=self.med_prop,
            state_inlet=self.state_inlet,
            state_outlet=self.state_outlet
        )

    def calc_alpha_liquid(self, transport_properties) -> float:
        """
        Calculate the liquid-phase heat transfer coefficient.

        Args:
            transport_properties: Transport properties for the liquid phase.

        Returns:
            float: The liquid-phase heat transfer coefficient.
        """
        return self._liquid_heat_transfer.calc(
            transport_properties=transport_properties,
            m_flow=self.m_flow
        )

    def calc_alpha_gas(self, transport_properties) -> float:
        """
        Calculate the gas-phase heat transfer coefficient.

        Args:
            transport_properties: Transport properties for the gas phase.

        Returns:
            float: The gas-phase heat transfer coefficient.
        """
        return self._gas_heat_transfer.calc(
            transport_properties=transport_properties,
            m_flow=self.m_flow
        )

    def calc_alpha_secondary(self, transport_properties) -> float:
        """
        Calculate the secondary-medium heat transfer coefficient.

        Args:
            transport_properties: Transport properties for the secondary medium.

        Returns:
            float: The secondary-medium heat transfer coefficient.
        """
        return self._secondary_heat_transfer.calc(
            transport_properties=transport_properties,
            m_flow=self.m_flow_secondary
        )

    def calc_wall_heat_transfer(self) -> float:
        """
        Calculate the heat transfer coefficient inside the wall.

        Returns:
            float: The wall heat transfer coefficient.
        """
        # Arguments are not required
        return self._wall_heat_transfer.calc(
            transport_properties=media.TransportProperties(),
            m_flow=0
        )

    @property
    def m_flow_secondary(self) -> float:
        return self._m_flow_secondary

    @m_flow_secondary.setter
    def m_flow_secondary(self, m_flow: float):
        self._m_flow_secondary = m_flow
        self._m_flow_secondary_cp = self._m_flow_secondary * self._secondary_cp

    @property
    def m_flow_secondary_cp(self):
        return self._m_flow_secondary_cp

    def calc_secondary_cp(self, T: float, p=None):
        """
        Calculate and set the heat capacity rate m_flow_cp of the secondary medium.

        Args:
            T (float): Temperature of the secondary medium.
            p (float, optional): Pressure of the secondary medium. Defaults to None.
        """
        self._secondary_cp = self.calc_transport_properties_secondary_medium(T=T, p=p).cp
        self._m_flow_secondary_cp = self.m_flow_secondary * self._secondary_cp

    def calc_secondary_Q_flow(self, Q_flow: float) -> float:
        return Q_flow

    def calc_Q_flow(self) -> float:
        """
        Calculate the total heat flow rate.

        Returns:
            float: The total heat flow rate.
        """
        return self.m_flow * abs(self.state_inlet.h - self.state_outlet.h)

    def calc_transport_properties_secondary_medium(self, T, p=None) -> media.TransportProperties:
        """
        Calculate the transport properties for the selected secondary_medium.

        Args:
            T (float): Temperature in K.
            p (float, optional): Pressure to use. Defaults to None.

        Returns:
            media.TransportProperties: The calculated transport properties.
        """
        if p is None:
            if self.secondary_medium == "water":
                p = 2e5  # 2 bar (default hydraulic pressure)
            elif self.secondary_medium == "air":
                p = 101325  # 1 atm
            else:
                raise NotImplementedError(
                    "Default pressures for secondary_mediums aside from water and air are not supported yet."
                )
        # Calc state
        state = self.med_prop_sec.calc_state("PT", p, T)
        # Return properties
        return self.med_prop_sec.calc_transport_properties(state)



"""
Module with models for pipe-to-wall heat transfer.
"""

from .heat_transfer import HeatTransfer, calc_reynolds_pipe
from vclibpy.media import TransportProperties


class TurbulentFluidInPipeToWallTransfer(HeatTransfer):
    """
    Class to model turbulent heat exchange in a pipe.

    Args:
        method (str):
            Equation to calc the nusselt number of turbulent flow for
            a given Re and Pr number.
            Note: Just for one-phase heat transfer!!
            Implemented Options are:

            - Taler2016
            - Domanski1989_sp_smooth
            - Amalfi2016
            - ScriptWSÜ. For turbulent regimes, eta_by_eta_w is assumed to be one.

            Refer to the paper / documents or the function in this class for more
            info on numbers and assumptions
        characteristic_length (float):
            Length to calculate the similitude approach for the
            heat transfer from ref -> wall. For heat pumps this is
            always the Diameter of the HE in m
    """

    def __init__(self, method: str, characteristic_length: float):
        self.method = method
        self.characteristic_length = characteristic_length

    def calc(self, transport_properties: TransportProperties, m_flow: float) -> float:
        """
        Calculate heat transfer coefficient from refrigerant to the wall of the heat exchanger.

        The flow is assumed to be always turbulent and is based on a calibrated
        Nusselt correlation.

        Args:
            transport_properties (TransportProperties): Transport properties of the fluid.
            m_flow (float): Mass flow rate of the fluid.

        Returns:
            float: Heat transfer coefficient from refrigerant to HE in W/(m^2*K).
        """
        Re = calc_reynolds_pipe(
            characteristic_length=self.characteristic_length,
            dynamic_viscosity=transport_properties.dyn_vis,
            m_flow=m_flow
        )
        Nu = self.calc_turbulent_tube_nusselt(Re, transport_properties.Pr)
        return Nu * transport_properties.lam / self.characteristic_length

    def calc_turbulent_tube_nusselt(self, Re, Pr) -> float:
        """
        Calculate the Nuseelt number for turbulent heat transfer
        in a tube (used for ref/water->Wall in the evaporator and condernser).

        Args:
            Re (float): Reynolds number.
            Pr (float): Prandtl number.

        Returns:
            float: Nusselt number.
        """
        if self.method == "Taler2016":
            if Re < 3e3 or Re > 1e6:
                raise ValueError(f"Given Re {Re} is outside of allowed bounds for method {self.method}")
            if 0.1 <= Pr <= 1:
                return 0.02155 * Re ** 0.8018 * Pr ** 0.7095
            elif 1 < Pr <= 3:
                return 0.02155 * Re ** 0.8018 * Pr ** 0.7095
            elif 3 < Pr <= 1000:
                return 0.02155 * Re ** 0.8018 * Pr ** 0.7095
            else:
                raise ValueError(f"Given Pr {Pr} is outside of allowed bounds for method {self.method}")
        elif self.method == "ScriptWSÜ":
            if Re < 3000 or Re > 1e5:
                raise ValueError(f"Given Re {Re} is outside of allowed bounds for method {self.method}")
            eta_by_eta_w = 1
            return 0.027 * Re ** 0.8 ** Pr ** (1 / 3) * eta_by_eta_w ** 0.14
        elif self.method == "Amalfi2016":
            if Re <= 700:
                Nu = (0.0295 * Pr - 0.115) * Re ** 0.954
            else:
                Nu = (1.760 * Pr - 5.391) * Re ** 0.262
            if Nu < 0:
                raise ValueError(f"Given Pr {Pr} is outside of allowed bounds for method {self.method}")
            return Nu
        elif self.method == "Domanski1989_sp_smooth":
            # According to Domanski 1989 for singular phase and smooth surfaces
            return 0.023 * Re ** 0.8 * Pr ** 0.4
        else:
            raise TypeError(f"Method {self.method} not supported!")



import abc
import logging

import numpy as np
from vclibpy.datamodels import FlowsheetState, Inputs
from vclibpy.components.heat_exchangers.ntu import BasicNTU
from vclibpy.media import ThermodynamicState

logger = logging.getLogger(__name__)


class MovingBoundaryNTU(BasicNTU, abc.ABC):
    """
    Moving boundary NTU based heat exchanger.

    See parent classe for arguments.
    """

    def separate_phases(self, state_max: ThermodynamicState, state_min: ThermodynamicState, p: float):
        """
        Separates a flow with possible phase changes into three parts:
        subcooling (sc), latent phase change (lat), and superheating (sh)
        at the given pressure.

        Args:
            state_max (ThermodynamicState): State with higher enthalpy.
            state_min (ThermodynamicState): State with lower enthalpy.
            p (float): Pressure of phase change.

        Returns:
            Tuple[float, float, float, ThermodynamicState, ThermodynamicState]:
                Q_sc: Heat for subcooling.
                Q_lat: Heat for latent phase change.
                Q_sh: Heat for superheating.
                state_q0: State at vapor quality 0 and the given pressure.
                state_q1: State at vapor quality 1 and the given pressure.
        """
        # Get relevant states:
        state_q0 = self.med_prop.calc_state("PQ", p, 0)
        state_q1 = self.med_prop.calc_state("PQ", p, 1)
        Q_sc = max(0.0,
                   min((state_q0.h - state_min.h),
                       (state_max.h - state_min.h))) * self.m_flow
        Q_lat = max(0.0,
                    (min(state_max.h, state_q1.h) -
                     max(state_min.h, state_q0.h))) * self.m_flow
        Q_sh = max(0.0,
                   min((state_max.h - state_q1.h),
                       (state_max.h - state_min.h))) * self.m_flow
        return Q_sc, Q_lat, Q_sh, state_q0, state_q1

    def iterate_area(self, dT_max, alpha_pri, alpha_sec, Q) -> float:
        """
        Iteratively calculates the required area for the heat exchange.

        Args:
            dT_max (float): Maximum temperature differential.
            alpha_pri (float): Heat transfer coefficient for the primary medium.
            alpha_sec (float): Heat transfer coefficient for the secondary medium.
            Q (float): Heat flow rate.

        Returns:
            float: Required area for heat exchange.
        """
        _accuracy = 1e-6  # square mm
        _step = 1.0
        R = self.calc_R()
        k = self.calc_k(alpha_pri, alpha_sec)
        m_flow_cp_min = self.calc_m_flow_cp_min()
        # First check if point is feasible at all
        if dT_max <= 0:
            return self.A
        eps_necessary = Q / (m_flow_cp_min * dT_max)

        # Special cases:
        # ---------------
        # eps is equal or higher than 1, an infinite amount of area would be necessary.
        if eps_necessary >= 1:
            return self.A
        # eps is lower or equal to zero: No Area required (Q<=0)
        if eps_necessary <= 0:
            return 0

        area = 0.0
        while True:
            NTU = self.calc_NTU(area, k, m_flow_cp_min)
            eps = self.calc_eps(R, NTU)
            if eps >= eps_necessary:
                if _step <= _accuracy:
                    break
                else:
                    # Go back
                    area -= _step
                    _step /= 10
                    continue
            if _step < _accuracy and area > self.A:
                break
            area += _step

        return min(area, self.A)


class MovingBoundaryNTUCondenser(MovingBoundaryNTU):
    """
    Condenser class which implements the actual `calc` method.

    Assumptions:
    - No phase changes in secondary medium
    - cp of secondary medium is constant over heat-exchanger

    See parent classes for arguments.
    """

    def calc(self, inputs: Inputs, fs_state: FlowsheetState) -> (float, float):
        """
        Calculate the heat exchanger with the NTU-Method based on the given inputs.

        The flowsheet state can be used to save important variables
        during calculation for later analysis.

        Both return values are used to check if the heat transfer is valid or not.

        Args:
            inputs (Inputs): The inputs for the calculation.
            fs_state (FlowsheetState): The flowsheet state to save important variables.

        Returns:
            Tuple[float, float]:
                error: Error in percentage between the required and calculated heat flow rates.
                dT_min: Minimal temperature difference (can be negative).
        """
        self.m_flow_secondary = inputs.m_flow_con  # [kg/s]
        self.calc_secondary_cp(T=inputs.T_con_in)

        # First we separate the flow:
        Q_sc, Q_lat, Q_sh, state_q0, state_q1 = self.separate_phases(
            self.state_inlet,
            self.state_outlet,
            self.state_inlet.p
        )
        Q = Q_sc + Q_lat + Q_sh

        # Note: As Q_con_ntu has to converge to Q_con (m_ref*delta_h), we can safely
        # calculate the output temperature.

        T_mean = inputs.T_con_in + self.calc_secondary_Q_flow(Q) / (self.m_flow_secondary_cp * 2)
        tra_prop_med = self.calc_transport_properties_secondary_medium(T_mean)
        alpha_med_wall = self.calc_alpha_secondary(tra_prop_med)

        # Calculate secondary_medium side temperatures:
        # Assumption loss is the same correlation for each regime
        T_sc = inputs.T_con_in + self.calc_secondary_Q_flow(Q_sc) / self.m_flow_secondary_cp
        T_sh = T_sc + self.calc_secondary_Q_flow(Q_lat) / self.m_flow_secondary_cp
        T_out = T_sh + self.calc_secondary_Q_flow(Q_sh) / self.m_flow_secondary_cp

        # 1. Regime: Subcooling
        Q_sc_ntu, A_sc = 0, 0
        if Q_sc > 0 and (state_q0.T != self.state_outlet.T):
            self.set_primary_cp((state_q0.h - self.state_outlet.h) / (state_q0.T - self.state_outlet.T))
            # Get transport properties:
            tra_prop_ref_con = self.med_prop.calc_mean_transport_properties(state_q0, self.state_outlet)
            alpha_ref_wall = self.calc_alpha_liquid(tra_prop_ref_con)

            # Only use still available area:
            A_sc = self.iterate_area(dT_max=(state_q0.T - inputs.T_con_in),
                                     alpha_pri=alpha_ref_wall,
                                     alpha_sec=alpha_med_wall,
                                     Q=Q_sc)
            A_sc = min(self.A, A_sc)

            Q_sc_ntu, k_sc = self.calc_Q_ntu(dT_max=(state_q0.T - inputs.T_con_in),
                                             alpha_pri=alpha_ref_wall,
                                             alpha_sec=alpha_med_wall,
                                             A=A_sc)

        # 2. Regime: Latent heat exchange
        Q_lat_ntu, A_lat = 0, 0
        if Q_lat > 0:
            self.set_primary_cp(np.inf)
            # Get transport properties:
            alpha_ref_wall = self.calc_alpha_two_phase(
                state_q0=state_q0,
                state_q1=state_q1,
                fs_state=fs_state,
                inputs=inputs
            )

            A_lat = self.iterate_area(dT_max=(state_q1.T - T_sc),
                                      alpha_pri=alpha_ref_wall,
                                      alpha_sec=alpha_med_wall,
                                      Q=Q_lat)
            # Only use still available area:
            A_lat = min(self.A - A_sc, A_lat)

            Q_lat_ntu, k_lat = self.calc_Q_ntu(dT_max=(state_q1.T - T_sc),
                                               alpha_pri=alpha_ref_wall,
                                               alpha_sec=alpha_med_wall,
                                               A=A_lat)
            logger.debug(f"con_lat: pri: {round(alpha_ref_wall, 2)} sec: {round(alpha_med_wall, 2)}")

        # 3. Regime: Superheat heat exchange
        Q_sh_ntu, A_sh = 0, 0
        if Q_sh and (self.state_inlet.T != state_q1.T):
            self.set_primary_cp((self.state_inlet.h - state_q1.h) / (self.state_inlet.T - state_q1.T))
            # Get transport properties:
            tra_prop_ref_con = self.med_prop.calc_mean_transport_properties(self.state_inlet, state_q1)
            alpha_ref_wall = self.calc_alpha_gas(tra_prop_ref_con)

            # Only use still available area:
            A_sh = self.A - A_sc - A_lat

            Q_sh_ntu, k_sh = self.calc_Q_ntu(dT_max=(self.state_inlet.T - T_sh),
                                             alpha_pri=alpha_ref_wall,
                                             alpha_sec=alpha_med_wall,
                                             A=A_sh)
            logger.debug(f"con_sh: pri: {round(alpha_ref_wall, 2)} sec: {round(alpha_med_wall, 2)}")

        Q_ntu = Q_sh_ntu + Q_sc_ntu + Q_lat_ntu
        error = (Q_ntu / Q - 1) * 100
        # Get possible dT_min:
        dT_min_in = self.state_outlet.T - inputs.T_con_in
        dT_min_out = self.state_inlet.T - T_out
        dT_min_LatSH = state_q1.T - T_sh

        fs_state.set(name="A_con_sh", value=A_sh, unit="m2", description="Area for superheat heat exchange in condenser")
        fs_state.set(name="A_con_lat", value=A_lat, unit="m2", description="Area for latent heat exchange in condenser")
        fs_state.set(name="A_con_sc", value=A_sc, unit="m2", description="Area for subcooling heat exchange in condenser")

        return error, min(dT_min_in,
                          dT_min_LatSH,
                          dT_min_out)


class MovingBoundaryNTUEvaporator(MovingBoundaryNTU):
    """
    Evaporator class which implements the actual `calc` method.

    Assumptions:
    - No phase changes in secondary medium
    - cp of secondary medium is constant over heat-exchanger

    See parent classes for arguments.
    """

    def calc(self, inputs: Inputs, fs_state: FlowsheetState) -> (float, float):
        """
        Calculate the heat exchanger with the NTU-Method based on the given inputs.

        The flowsheet state can be used to save important variables
        during calculation for later analysis.

        Both return values are used to check if the heat transfer is valid or not.

        Args:
            inputs (Inputs): The inputs for the calculation.
            fs_state (FlowsheetState): The flowsheet state to save important variables.

        Returns:
            Tuple[float, float]:
                error: Error in percentage between the required and calculated heat flow rates.
                dT_min: Minimal temperature difference (can be negative).
        """
        self.m_flow_secondary = inputs.m_flow_eva  # [kg/s]
        self.calc_secondary_cp(T=inputs.T_eva_in)

        # First we separate the flow:
        Q_sc, Q_lat, Q_sh, state_q0, state_q1 = self.separate_phases(
            self.state_outlet,
            self.state_inlet,
            self.state_inlet.p
        )

        Q = Q_sc + Q_lat + Q_sh

        # Note: As Q_eva_ntu has to converge to Q_eva (m_ref*delta_h), we can safely
        # calculate the output temperature.
        T_mean = inputs.T_eva_in - Q / (self.m_flow_secondary_cp * 2)
        tra_prop_med = self.calc_transport_properties_secondary_medium(T_mean)
        alpha_med_wall = self.calc_alpha_secondary(tra_prop_med)

        # Calculate secondary_medium side temperatures:
        T_sh = inputs.T_eva_in - Q_sh / self.m_flow_secondary_cp
        T_sc = T_sh - Q_lat / self.m_flow_secondary_cp
        T_out = T_sc - Q_sc / self.m_flow_secondary_cp

        # 1. Regime: Superheating
        Q_sh_ntu, A_sh = 0, 0
        if Q_sh and (self.state_outlet.T != state_q1.T):
            self.set_primary_cp((self.state_outlet.h - state_q1.h) / (self.state_outlet.T - state_q1.T))
            # Get transport properties:
            tra_prop_ref_eva = self.med_prop.calc_mean_transport_properties(self.state_outlet, state_q1)
            alpha_ref_wall = self.calc_alpha_gas(tra_prop_ref_eva)

            if Q_lat > 0:
                A_sh = self.iterate_area(dT_max=(inputs.T_eva_in - state_q1.T),
                                         alpha_pri=alpha_ref_wall,
                                         alpha_sec=alpha_med_wall,
                                         Q=Q_sh)
            else:
                # if only sh is present --> full area:
                A_sh = self.A

            # Only use still available area
            A_sh = min(self.A, A_sh)

            Q_sh_ntu, k_sh = self.calc_Q_ntu(dT_max=(inputs.T_eva_in - state_q1.T),
                                             alpha_pri=alpha_ref_wall,
                                             alpha_sec=alpha_med_wall,
                                             A=A_sh)

            logger.debug(f"eva_sh: pri: {round(alpha_ref_wall, 2)} sec: {round(alpha_med_wall, 2)}")

        # 2. Regime: Latent heat exchange
        Q_lat_ntu, A_lat = 0, 0
        if Q_lat > 0:
            self.set_primary_cp(np.inf)

            alpha_ref_wall = self.calc_alpha_two_phase(
                state_q0=state_q0,
                state_q1=state_q1,
                fs_state=fs_state,
                inputs=inputs
            )

            if Q_sc > 0:
                A_lat = self.iterate_area(dT_max=(T_sh - self.state_inlet.T),
                                          alpha_pri=alpha_ref_wall,
                                          alpha_sec=alpha_med_wall,
                                          Q=Q_lat)
            else:
                A_lat = self.A - A_sh

            # Only use still available area:
            A_lat = min(self.A - A_sh, A_lat)
            Q_lat_ntu, k_lat = self.calc_Q_ntu(dT_max=(T_sh - self.state_inlet.T),
                                               alpha_pri=alpha_ref_wall,
                                               alpha_sec=alpha_med_wall,
                                               A=A_lat)
            logger.debug(f"eva_lat: pri: {round(alpha_ref_wall, 2)} sec: {round(alpha_med_wall, 2)}")

        # 3. Regime: Subcooling
        Q_sc_ntu, A_sc = 0, 0
        if Q_sc > 0 and (state_q0.T != self.state_inlet.T):
            self.set_primary_cp((state_q0.h - self.state_inlet.h) / (state_q0.T - self.state_inlet.T))
            # Get transport properties:
            tra_prop_ref_eva = self.med_prop.calc_mean_transport_properties(state_q0, self.state_inlet)
            alpha_ref_wall = self.calc_alpha_liquid(tra_prop_ref_eva)

            # Only use still available area:
            A_sc = self.A - A_sh - A_lat

            Q_sc_ntu, k_sc = self.calc_Q_ntu(dT_max=(T_sc - self.state_inlet.T),
                                             alpha_pri=alpha_ref_wall,
                                             alpha_sec=alpha_med_wall,
                                             A=A_sc)

        Q_ntu = Q_sh_ntu + Q_sc_ntu + Q_lat_ntu
        error = (Q_ntu / Q - 1) * 100
        # Get dT_min
        dT_min_in = inputs.T_eva_in - self.state_outlet.T
        dT_min_out = T_out - self.state_inlet.T

        fs_state.set(name="A_eva_sh", value=A_sh, unit="m2", description="Area for superheat heat exchange in evaporator")
        fs_state.set(name="A_eva_lat", value=A_lat, unit="m2", description="Area for latent heat exchange in evaporator")

        return error, min(dT_min_out, dT_min_in)



import abc
import logging
from copy import deepcopy
import numpy as np

from vclibpy.flowsheets import BaseCycle
from vclibpy.datamodels import Inputs, FlowsheetState
from vclibpy.components.compressors import Compressor
from vclibpy.components.expansion_valves import ExpansionValve
from vclibpy.media import ThermodynamicState

logger = logging.getLogger(__name__)


class BaseVaporInjection(BaseCycle, abc.ABC):
    """
    Partial cycle with vapor injection, using
    two separated compressors and expansion valves.

    Notes
    -----
    See parent docstring for info on further assumptions and parameters.
    """

    flowsheet_name = "VaporInjectionPhaseSeparator"

    def __init__(
            self,
            high_pressure_compressor: Compressor,
            low_pressure_compressor: Compressor,
            high_pressure_valve: ExpansionValve,
            low_pressure_valve: ExpansionValve,
            **kwargs):
        super().__init__(**kwargs)
        self.high_pressure_compressor = high_pressure_compressor
        self.low_pressure_compressor = low_pressure_compressor
        self.high_pressure_valve = high_pressure_valve
        self.low_pressure_valve = low_pressure_valve
        # Avoid nasty bugs for setting states
        if id(high_pressure_compressor) == id(low_pressure_compressor):
            self.high_pressure_compressor = deepcopy(low_pressure_compressor)
        if id(low_pressure_valve) == id(high_pressure_valve):
            self.high_pressure_valve = deepcopy(low_pressure_valve)

    def get_all_components(self):
        return super().get_all_components() + [
            self.high_pressure_compressor,
            self.low_pressure_compressor,
            self.high_pressure_valve,
            self.low_pressure_valve,
        ]

    def calc_states(self, p_1, p_2, inputs: Inputs, fs_state: FlowsheetState):
        k_vapor_injection = inputs.get("k_vapor_injection", default=1)
        # Default according to Xu, 2019

        p_vapor_injection = k_vapor_injection * np.sqrt(p_1 * p_2)

        # Condenser outlet
        self.set_condenser_outlet_based_on_subcooling(p_con=p_2, inputs=inputs)
        # High pressure EV
        self.high_pressure_valve.state_inlet = self.condenser.state_outlet
        self.high_pressure_valve.calc_outlet(p_outlet=p_vapor_injection)

        # Calculate low compressor stage to already have access to the mass flow rates.
        self.set_evaporator_outlet_based_on_superheating(p_eva=p_1, inputs=inputs)
        self.low_pressure_compressor.state_inlet = self.evaporator.state_outlet
        self.low_pressure_compressor.calc_state_outlet(
            p_outlet=p_vapor_injection, inputs=inputs, fs_state=fs_state
        )
        m_flow_low = self.low_pressure_compressor.calc_m_flow(inputs=inputs, fs_state=fs_state)
        self.evaporator.m_flow = self.low_pressure_compressor.m_flow

        # Injection component:
        x_vapor_injection, h_vapor_injection, state_low_ev_inlet = self.calc_injection()

        # Low pressure EV
        self.low_pressure_valve.state_inlet = state_low_ev_inlet
        self.low_pressure_valve.calc_outlet(p_outlet=p_1)
        # Evaporator
        self.evaporator.state_inlet = self.low_pressure_valve.state_outlet

        # Ideal Mixing of state_5 and state_1_VI:
        h_1_VI_mixed = (
                (1-x_vapor_injection) * self.low_pressure_compressor.state_outlet.h +
                x_vapor_injection * h_vapor_injection
        )
        self.high_pressure_compressor.state_inlet = self.med_prop.calc_state(
            "PH", p_vapor_injection, h_1_VI_mixed
        )
        self.high_pressure_compressor.calc_state_outlet(
            p_outlet=p_2, inputs=inputs, fs_state=fs_state
        )

        # Check m_flow of both compressor stages to check if
        # there would be an asymmetry of how much refrigerant is transported
        m_flow_high = self.high_pressure_compressor.calc_m_flow(
            inputs=inputs, fs_state=fs_state
        )
        m_flow_low_should = m_flow_high * (1-x_vapor_injection)
        percent_deviation = (m_flow_low - m_flow_low_should) / m_flow_low_should * 100
        logger.debug("Deviation of mass flow rates is %s percent", percent_deviation)

        # Set states
        self.condenser.m_flow = self.high_pressure_compressor.m_flow
        self.condenser.state_inlet = self.high_pressure_compressor.state_outlet

        fs_state.set(
            name="T_1", value=self.evaporator.state_outlet.T,
            unit="K", description="Refrigerant temperature at evaporator outlet"
        )
        fs_state.set(
            name="T_2", value=self.high_pressure_compressor.state_outlet.T,
            unit="K", description="Compressor outlet temperature"
        )
        fs_state.set(
            name="T_3", value=self.condenser.state_outlet.T,
            unit="K", description="Refrigerant temperature at condenser outlet"
        )
        fs_state.set(
            name="T_4", value=self.evaporator.state_inlet.T,
            unit="K", description="Refrigerant temperature at evaporator inlet"
        )
        fs_state.set(
            name="p_con", value=p_2,
            unit="Pa", description="Condensation pressure"
        )
        fs_state.set(
            name="p_eva", value=p_1,
            unit="Pa", description="Evaporation pressure"
        )

    def calc_injection(self) -> (float, float, ThermodynamicState):
        """
        Calculate the injection component, e.g. phase separator
        or heat exchanger.
        In this function, child classes must set inlets
        and calculate outlets of additional components.

        Returns:
            float: Portion of vapor injected (x)
            float: Enthalpy of vapor injected
            ThermodynamicState: Inlet state of low pressure expansion valve
        """
        raise NotImplementedError

    def calc_electrical_power(self, inputs: Inputs, fs_state: FlowsheetState):
        P_el_low = self.low_pressure_compressor.calc_electrical_power(
            inputs=inputs, fs_state=fs_state
        )
        P_el_high = self.high_pressure_compressor.calc_electrical_power(
            inputs=inputs, fs_state=fs_state
        )
        fs_state.set(
            name="P_el_low",
            value=P_el_low,
            unit="W",
            description="Electrical power consumption of low stage compressor"
        )
        fs_state.set(
            name="P_el_high",
            value=P_el_high,
            unit="W",
            description="Electrical power consumption of high stage compressor"
        )
        return P_el_low + P_el_high

    def get_states_in_order_for_plotting(self):
        """
        List with all relevant states of two-stage cycle
        except the intermediate component, e.g. phase separator
        or heat exchanger.
        """
        return [
            self.low_pressure_valve.state_inlet,
            self.low_pressure_valve.state_outlet,
            self.evaporator.state_inlet,
            self.med_prop.calc_state("PQ", self.evaporator.state_inlet.p, 1),
            self.evaporator.state_outlet,
            self.low_pressure_compressor.state_inlet,
            self.low_pressure_compressor.state_outlet,
            self.high_pressure_compressor.state_inlet,
            self.high_pressure_compressor.state_outlet,
            self.condenser.state_inlet,
            self.med_prop.calc_state("PQ", self.condenser.state_inlet.p, 1),
            self.med_prop.calc_state("PQ", self.condenser.state_inlet.p, 0),
            self.condenser.state_outlet,
            self.high_pressure_valve.state_inlet,
            self.high_pressure_valve.state_outlet,
        ]



"""
Module with a simple phase separator model.
"""

from vclibpy.media import ThermodynamicState
from vclibpy.components.component import BaseComponent


class PhaseSeparator(BaseComponent):
    """
    A simple phase separator model.
    """

    def __init__(self):
        super().__init__()
        self._state_outlet_liquid: ThermodynamicState = None
        self._state_outlet_vapor: ThermodynamicState = None

    @BaseComponent.state_inlet.setter
    def state_inlet(self, state_inlet: ThermodynamicState):
        """
        Set the state of the inlet and calculate the outlet states for liquid and vapor phases.

        Args:
            state_inlet (ThermodynamicState): Inlet state.
        """
        self._state_inlet = state_inlet
        self.state_outlet_vapor = self.med_prop.calc_state("PQ", self.state_inlet.p, 1)
        self.state_outlet_liquid = self.med_prop.calc_state("PQ", self.state_inlet.p, 0)

    @BaseComponent.state_outlet.setter
    def state_outlet(self, state: ThermodynamicState):
        """
        This outlet is disabled for this component.

        Args:
            state (ThermodynamicState): Outlet state.
        """
        raise NotImplementedError("This outlet is disabled for this component")

    @property
    def state_outlet_vapor(self) -> ThermodynamicState:
        """
        Getter for the outlet state of the vapor phase.

        Returns:
            ThermodynamicState: Outlet state for the vapor phase.
        """
        return self._state_outlet_vapor

    @state_outlet_vapor.setter
    def state_outlet_vapor(self, state: ThermodynamicState):
        """
        Setter for the outlet state of the vapor phase.

        Args:
            state (ThermodynamicState): Outlet state for the vapor phase.
        """
        self._state_outlet_vapor = state

    @property
    def state_outlet_liquid(self) -> ThermodynamicState:
        """
        Getter for the outlet state of the liquid phase.

        Returns:
            ThermodynamicState: Outlet state for the liquid phase.
        """
        return self._state_outlet_liquid

    @state_outlet_liquid.setter
    def state_outlet_liquid(self, state: ThermodynamicState):
        """
        Setter for the outlet state of the liquid phase.

        Args:
            state (ThermodynamicState): Outlet state for the liquid phase.
        """
        self._state_outlet_liquid = state



"""
Module with simple bernoulli expansion valve
"""
from vclibpy.components.expansion_valves import ExpansionValve


class Bernoulli(ExpansionValve):
    """
    Simple Bernoulli model.

    Args:
        A (float): Cross-sectional area of the expansion valve.
    """

    def calc_m_flow_at_opening(self, opening):
        return opening * self.A * (2 * self.state_inlet.d * (self.state_inlet.p - self.state_outlet.p)) ** 0.5

    def calc_opening_at_m_flow(self, m_flow, **kwargs):
        return (
                m_flow /
                (self.A * (2 * self.state_inlet.d * (self.state_inlet.p - self.state_outlet.p)) ** 0.5)
        )



import itertools
import pathlib

import pandas as pd
import sdf


def save_to_sdf(data: dict, save_path: pathlib.Path):
    """
    Save given input dictionary to a sdf file in the given save_path

    Args:
        data (dict):
            A dictionary with the following structure:
            Keys: Flowsheet_name
            Values: A dictionary with the following structure:

            - Keys: Fluids
            - Values: Data dictionaries with the following structure:
                A tuple with three values, in that order:

                - scales: Of the given nd_data, e.g. T_Amb, n
                - nd_data: More-dimensional data, e.g. COP
                - parameters: Scalar values, like m_flow_con or similar

        save_path (pathlib.Path): Where to store the data
        flowsheet_name (str): Name of the flowsheet. This is the top level group
    :return:
    """
    if isinstance(save_path, str):
        save_path = pathlib.Path(save_path)
    _all_groups = []

    for flowsheet_name, fluid_dict in data.items():
        _all_fluids = []
        for fluid, fluid_data in fluid_dict.items():
            # First write scales
            _scales = []
            for scale_name, scale_values in fluid_data[0].items():
                _scales.append(sdf.Dataset(scale_name,
                                           data=scale_values["data"],
                                           unit=scale_values["unit"],
                                           is_scale=True,
                                           display_name=scale_name,
                                           comment=scale_values.get("comment", "")))
            # Now the ND-Data:
            _nd_data = []
            for data_name, data_values in fluid_data[1].items():
                _nd_data.append(sdf.Dataset(data_name,
                                            data=data_values["data"],
                                            unit=data_values["unit"],
                                            scales=_scales,
                                            comment=data_values.get("comment", "")))
            # Now the constant parameters
            _paras = []
            for para_name, para_value in fluid_data[2].items():
                _paras.append(sdf.Dataset(para_name,
                                          data=para_value["data"],
                                          unit=para_value["unit"],
                                          comment=para_value.get("comment", "")))

            # Save everything
            fluid_group = sdf.Group(fluid, comment="Values for fluid", datasets=_scales + _nd_data + _paras)
            _all_fluids.append(fluid_group)

        flowsheet_group = sdf.Group(flowsheet_name,
                                    comment="Multiple fluids for the flowsheet",
                                    groups=_all_fluids)
        _all_groups.append(flowsheet_group)

    parent_group = sdf.Group("/", comment="Generated with VCLibPy", groups=_all_groups)
    sdf.save(save_path, group=parent_group)
    return save_path


def merge_sdfs(filepaths, save_path):
    """
    Merge given files and return a merged file.
    Be careful if both files contain the same combination.
    Then, the latter element of the list will overwrite the first one.

    Args:
        filepaths (list): List with paths to the files
        save_path (str): Save path for the new file
    """
    _all_flowsheets = {}
    # Merge to structure
    for fpath in filepaths:
        dataset = sdf.load(fpath)
        for flowsheet in dataset.groups:
            fluid_data = {fluid.name: fluid for fluid in flowsheet.groups}
            if flowsheet.name not in _all_flowsheets:
                _all_flowsheets.update({flowsheet.name: fluid_data})
            else:
                _all_flowsheets[flowsheet.name].update(fluid_data)

    # Write structure
    _all_groups = []
    for flowsheet_name, fluid_dict in _all_flowsheets.items():
        _all_fluids = []
        for fluid, data in fluid_dict.items():
            _all_fluids.append(data)
        flowsheet_group = sdf.Group(flowsheet_name,
                                    comment="Multiple fluids for the flowsheet",
                                    groups=_all_fluids)
        _all_groups.append(flowsheet_group)

    parent_group = sdf.Group("/", comment="Generated with python script", groups=_all_groups)
    sdf.save(save_path, group=parent_group)
    return save_path


def sdf_to_csv(filepath: pathlib.Path, save_path: pathlib.Path):
    """
    Convert a given .sdf file to multiple excel files,
    for each combination of flowsheet and refrigerant one file.

    Args:
        filepath (pathlib.Path): sdf file
        save_path (pathlib.Path): Directory where to store the csv files.
    """
    dataset = sdf.load(str(filepath))
    for flowsheet in dataset.groups:
        for fluid in flowsheet.groups:
            dfs = []
            for data in fluid.datasets:
                if _is_nd(data):
                    dfs.append(_unpack_nd_data(data))
            df = pd.concat(dfs, axis=1)
            df = df.loc[:, ~df.columns.duplicated()]
            df.to_csv(save_path.joinpath(f"{flowsheet.name}_{fluid.name}.csv"))


def _get_name(data):
    return f"{data.name} in {data.unit} ({data.comment})"


def _is_nd(data):
    if data.scales == [None]:
        return False
    return True


def _unpack_nd_data(data):
    column_name = _get_name(data)
    scale_names = [_get_name(scale) for scale in data.scales]
    scales_with_idx = [[(idx, value) for idx, value in enumerate(scale.data)] for scale in data.scales]
    all_data = []
    for scales in itertools.product(*scales_with_idx):
        indexer = tuple([scale[0] for scale in scales])
        values = [scale[1] for scale in scales]
        all_data.append({
            column_name: data.data[indexer],
            **{name: value for name, value in zip(scale_names, values)}
        })
    return pd.DataFrame(all_data)


if __name__ == '__main__':
    sdf_to_csv(
        filepath=pathlib.Path(r"D:\00_temp\calibration_jmo\Optihorst_3D_vclibpy.sdf"),
        save_path=pathlib.Path(r"D:\04_git\vclibpy\tests\regression_data")
    )



"""
Module with constant heat transfer assumptions
"""
import abc

from vclibpy.media import TransportProperties


class ConstantHeatTransfer(abc.ABC):
    """
    Constant heat transfer assumption

    Args:
        alpha (float):
            Constant heat transfer coefficient in W/(m2*K)
    """

    def __init__(self, alpha: float):
        self.alpha = alpha

    def calc(self, transport_properties: TransportProperties, m_flow: float) -> float:
        """
        Calculate constant heat transfer coefficient.

        Args:
            transport_properties (TransportProperties): Transport properties of the medium (not used).
            m_flow (float): Mass flow rate (not used).

        Returns:
            float: Constant heat transfer coefficient in W/(m2*K).

        """
        return self.alpha


class ConstantTwoPhaseHeatTransfer(abc.ABC):
    """
    Constant heat transfer assumption.

    Args:
        alpha (float):
            Constant heat transfer coefficient in W/(m2*K).
    """

    def __init__(self, alpha: float):
        self.alpha = alpha

    def calc(self, **kwargs) -> float:
        """
        Calculate constant two-phase heat transfer coefficient.

        Args:
            **kwargs: Allows to set arguments for different heat transfer classes which are not used here.

        Returns:
            float: Constant two-phase heat transfer coefficient in W/(m2*K).

        """
        return self.alpha



"""
Functions to generate HP Maps automatically
"""
import logging
import pathlib
import os
from typing import List, Union
import multiprocessing
import numpy as np
import pandas as pd
from vclibpy.datamodels import FlowsheetState, Inputs
from vclibpy.flowsheets import BaseCycle
from vclibpy import utils

logger = logging.getLogger(__name__)


def calc_multiple_states(
        save_path: pathlib.Path,
        heat_pump: BaseCycle,
        inputs: List[Inputs],
        **kwargs):
    """
    Function to calculate the flowsheet states for all given inputs.
    All results are stored as a .xlsx file in the given save-path

    Args:
        save_path (pathlib.Path): Location where to save the results as xlsx.
        heat_pump (BaseCycle): A valid flowsheet
        inputs (List[Inputs]): A list with all inputs to simulate
        **kwargs: Solver settings for the flowsheet
    """
    rel_infos = []
    for i, single_inputs in enumerate(inputs):
        fs_state = None
        logger.info(f"Running combination {i+1}/{len(inputs)}.")
        try:
            fs_state = heat_pump.calc_steady_state(inputs=single_inputs,
                                                   **kwargs)
        except Exception as e:
            # Avoid loss of data if un-excepted errors occur.
            logger.error(f"An error occurred: {e}")
        if fs_state is None:
            fs_state = FlowsheetState()
        hp_state_dic = {
            **single_inputs.convert_to_str_value_format(with_unit_and_description=True),
            **fs_state.convert_to_str_value_format(with_unit_and_description=True)
        }
        rel_infos.append(hp_state_dic)
    df = pd.DataFrame(rel_infos)
    df.index.name = "State Number"
    df.to_excel(save_path.joinpath(f"{heat_pump}_{heat_pump.fluid}.xlsx"), sheet_name="HP_states", float_format="%.5f")


def full_factorial_map_generation(
        heat_pump: BaseCycle,
        T_eva_in_ar: Union[list, np.ndarray],
        T_con_in_ar: Union[list, np.ndarray],
        n_ar: Union[list, np.ndarray],
        m_flow_con: float,
        m_flow_eva: float,
        save_path: Union[pathlib.Path, str],
        dT_eva_superheating=5,
        dT_con_subcooling=0,
        use_multiprocessing: bool = False,
        save_plots: bool = False,
        **kwargs
) -> (pathlib.Path, pathlib.Path):
    """
    Run a full-factorial simulation to create performance maps
    used in other simulation tools like Modelica or to analyze
    the off-design of the flowsheet.
    The results are stored and returned as .sdf and .csv files.
    Currently, only varying T_eva_in, T_con_in, and n is implemented.
    However, changing this to more dimensions or other variables
    is not much work. In this case, please raise an issue.

    Args:
        heat_pump (BaseCycle): The flowsheet to use
        T_eva_in_ar: Array with inputs for T_eva_in
        T_con_in_ar: Array with inputs for T_con_in
        n_ar: Array with inputs for n_ar
        m_flow_con: Condenser mass flow rate
        m_flow_eva: Evaporator mass flow rate
        save_path: Where to save all results.
        dT_eva_superheating: Evaporator superheating
        dT_con_subcooling: Condenser subcooling
        use_multiprocessing:
            True to use multiprocessing. May speed up the calculation. Default is False
        save_plots:
            True to save plots of each steady state point. Default is False
        **kwargs: Solver settings for the flowsheet

    Returns:
        tuple (pathlib.Path, pathlib.Path):
            Path to the created .sdf file and to the .csv file
    """
    if isinstance(save_path, str):
        save_path = pathlib.Path(save_path)
    if save_plots:
        kwargs["save_path_plots"] = pathlib.Path(save_path).joinpath(f"plots_{heat_pump.flowsheet_name}_{heat_pump.fluid}")
        os.makedirs(kwargs["save_path_plots"], exist_ok=True)

    list_mp_inputs = []
    list_inputs = []
    idx_for_access_later = []
    for i_T_eva_in, T_eva_in in enumerate(T_eva_in_ar):
        for i_n, n in enumerate(n_ar):
            for i_T_con_in, T_con_in in enumerate(T_con_in_ar):
                idx_for_access_later.append([i_n, i_T_con_in, i_T_eva_in])
                inputs = Inputs(n=n,
                                T_eva_in=T_eva_in,
                                T_con_in=T_con_in,
                                m_flow_eva=m_flow_eva,
                                m_flow_con=m_flow_con,
                                dT_eva_superheating=dT_eva_superheating,
                                dT_con_subcooling=dT_con_subcooling)
                list_mp_inputs.append([heat_pump, inputs, kwargs])
                list_inputs.append(inputs)
    fs_states = []
    i = 0
    if use_multiprocessing:
        pool = multiprocessing.Pool(processes=multiprocessing.cpu_count())
        for fs_state in pool.imap(_calc_single_hp_state, list_mp_inputs):
            fs_states.append(fs_state)
            i += 1
            logger.info(f"Ran {i} of {len(list_mp_inputs)} points")
    else:
        for inputs in list_inputs:
            fs_state = _calc_single_hp_state([heat_pump, inputs, kwargs])
            fs_states.append(fs_state)
            i += 1
            logger.info(f"Ran {i} of {len(list_mp_inputs)} points")

    # Save to sdf
    result_shape = (len(n_ar), len(T_con_in_ar), len(T_eva_in_ar))
    _dummy = np.zeros(result_shape)  # Use a copy to avoid overwriting of values of any sort.
    _dummy[:] = np.nan
    # Get all possible values:
    all_variables = {}
    all_variables_info = {}
    variables_to_excel = []
    for fs_state, inputs in zip(fs_states, list_inputs):
        all_variables.update({var: _dummy.copy() for var in fs_state.get_variable_names()})
        all_variables_info.update({var: variable for var, variable in fs_state.get_variables().items()})
        variables_to_excel.append({
            **inputs.convert_to_str_value_format(with_unit_and_description=True),
            **fs_state.convert_to_str_value_format(with_unit_and_description=True),
        })

    # Save to excel
    save_path_sdf = save_path.joinpath(f"{heat_pump.flowsheet_name}_{heat_pump.fluid}.sdf")
    save_path_csv = save_path.joinpath(f"{heat_pump.flowsheet_name}_{heat_pump.fluid}.csv")
    pd.DataFrame(variables_to_excel).to_csv(
        save_path_csv
    )

    for fs_state, idx_triple in zip(fs_states, idx_for_access_later):
        i_n, i_T_con_in, i_T_eva_in = idx_triple
        for variable_name, variable in fs_state.get_variables().items():
            all_variables[variable_name][i_n][i_T_con_in][i_T_eva_in] = variable.value

    _nd_data = {}
    for variable, nd_data in all_variables.items():
        _nd_data.update({
            variable: {
                "data": nd_data,
                "unit": all_variables_info[variable].unit,
                "comment": all_variables_info[variable].description}
        })

    _scale_values = {
        "n": n_ar,
        "T_con_in": T_con_in_ar,
        "T_eva_in": T_eva_in_ar
    }
    inputs: Inputs = list_inputs[0]
    _parameters = {}
    for name, variable in inputs.items():
        if name not in _scale_values:
            _parameters[name] = {
                "data": variable.value,
                "unit": variable.unit,
                "comment": variable.description
            }
    _scales = {}
    for name, data in _scale_values.items():
        _scales[name] = {
            "data": data,
            "unit": inputs.get(name).unit,
            "comment": inputs.get(name).description
        }

    sdf_data = {
        heat_pump.flowsheet_name:
            {
                heat_pump.fluid: (_scales, _nd_data, _parameters)
            }
    }
    utils.save_to_sdf(data=sdf_data, save_path=save_path_sdf)

    # Terminate heat pump med-props:
    heat_pump.terminate()

    return save_path_sdf, save_path_csv


def _calc_single_hp_state(data):
    """Helper function for a single state to enable multiprocessing"""
    heat_pump, inputs, kwargs = data
    fs_state = None
    try:
        fs_state = heat_pump.calc_steady_state(inputs=inputs,
                                               **kwargs)
    except Exception as e:
        logger.error(f"An error occurred for input: {inputs.__dict__}: {e}")
    if fs_state is None:
        fs_state = FlowsheetState()
    # Append the data to the dataframe
    return fs_state



"""Module with wrappers to access and handle media property databases.

This module provides interfaces to load media properties using various wrappers and
handle calculations related to media properties.

Classes:
    MedProp: Base class for all media property interfaces.

Functions:
    get_two_phase_limits: Return the states of the boundaries of the two-phase section for a given fluid.

"""
import abc
import logging
import warnings
from typing import List
import numpy as np

from vclibpy.media import ThermodynamicState, TransportProperties


logger = logging.getLogger(__name__)


class MedProp(abc.ABC):
    """Base class for all media property interfaces.

    This class serves as the base for defining interfaces to access and compute media properties.

    Methods:
        calc_state: Calculate the thermodynamic state based on mode and state variables.
        calc_transport_properties: Calculate transport properties for a given state.
        get_critical_point: Retrieve critical point information.
        get_molar_mass: Retrieve molar mass information.
        get_two_phase_limits: Retrieve the two-phase limits for plotting.
        calc_mean_transport_properties: Calculate the average transport properties for given states.
    """
    _fluid_mapper = {}

    def __init__(self, fluid_name):
        """Initialize the MedProp class instance.

        Args:
            fluid_name (str): The name of the fluid.
        """
        # Check if better internal names exist (e.g. air is modelled as air.ppf)
        self.fluid_name = self._fluid_mapper.get(fluid_name, fluid_name)
        self._two_phase_limits: dict = None

    def calc_state(self, mode: str, var1: float, var2: float):
        """Calculate the thermodynamic state based on the specified mode and state variables.

        This function calculates the thermodynamic state based on the chosen mode and provided state variables.
        The input state variables need to be in SI units.

        Notes:
            - PT does not work when the state might fall within the two-phase region.
            - Only functions for density are implemented. In cases where you know the specific volume, use the density
              functions with the inverse value.
            - Quality (q) may have values outside the 'physical' scope:
                - q = -998: Subcooled liquid
                - q = 998: Superheated vapor
                - q = 999: Supercritical state

        Possible modes include:
            - "PD": Pressure, Density
            - "PH": Pressure, Enthalpy
            - "PQ": Pressure, Quality
            - "PS": Pressure, Entropy
            - "PT": Pressure, Temperature
            - "PU": Pressure, Internal Energy
            - "TD": Temperature, Density
            - "TH": Temperature, Enthalpy
            - "TQ": Temperature, Quality
            - "TS": Temperature, Entropy
            - "TU": Temperature, Internal Energy
            - "DH": Density, Enthalpy
            - "DS": Density, Entropy
            - "DU": Density, Internal Energy

        Args:
            mode (str): Defines the given input state variables (see possible modes above).
            var1 (float): Value of the first state variable (as specified in the mode) in SI units.
            var2 (float): Value of the second state variable (as specified in the mode) in SI units.

        Returns:
            ThermodynamicState: A ThermodynamicState instance with state variables.

        Raises:
            AssertionError: If the given mode is not within the available options.
        """
        available_options = ['PD', 'PH', 'PQ', 'PS', 'PT',
                             'PU', 'TD', 'TH', 'TQ', 'TS',
                             'TU', 'DH', 'DS', 'DU', ]
        assert mode in available_options, f'Given mode {mode} is not in available options'

    def terminate(self):
        """
        Terminate the class.
        Default behaviour does nothing.
        """
        pass

    @abc.abstractmethod
    def calc_transport_properties(self, state: ThermodynamicState):
        """Calculate the transport properties for the given state.

        Args:
            state (ThermodynamicState): The current thermodynamic state.

        Returns:
            TransportProperties: An instance of TransportProperties.
        """
        pass

    @abc.abstractmethod
    def get_critical_point(self):
        """Retrieve critical point information for the fluid.

        Returns:
            Tuple[float, float, float]: A tuple containing critical point information
            (Temperature Tc [K], Pressure pc [Pa], Density dc [kg/m^3]).
        """
        pass

    @abc.abstractmethod
    def get_molar_mass(self):
        """Retrieve the molar mass of the current fluid.

        Returns:
            float: The molar mass M of the current fluid in kg/mol.
        """
        pass

    def get_two_phase_limits(self, quantity: str, p_min: int = 100000, p_step: int = 5000):
        """
        Retrieve the two-phase limits for plotting a specified quantity.

        This method returns the two-phase limits for a specified quantity (T, h, s, or p) in an array used for
        plotting purposes. It calculates the limits within the pressure range from p_min and quality (q) 0 to the
        critical pressure (pc), and then from the critical pressure to the pressure p_min and quality 1.

        Args:
            quantity (str): The specified quantity (T, h, s, or p).
            p_min (int, optional): The minimum pressure value to start iteration. Default is 100000 Pa.
            p_step (int, optional): The step size for pressure variation. Default is 5000 Pa.

        Returns:
            numpy.ndarray: An array containing the two-phase limits for the specified quantity.

        Raises:
            ValueError: If the given quantity is not supported (T, h, s, or p).
        """
        if self._two_phase_limits is not None:
            # Check existing two-phase limits
            p_min_old = self._two_phase_limits['p'][0]
            p_step_old = self._two_phase_limits['p'][1] - p_min_old
            if not np.isclose(p_min_old, p_min, 0, 10) or not np.isclose(p_step_old, p_step, 0, 10):
                warnings.warn(f"Overwriting previously calculated two-phase limits with "
                              f"p_min={p_min_old} and p_step={p_step_old}. This might take a few seconds.\n"
                              f"The quantity might not match with the previously calculated quantities.")
                self._two_phase_limits = None

        if self._two_phase_limits is None:
            # Calculate new two-phase limits for plotting
            _two_phase_limits = get_two_phase_limits(self, p_step=p_step, p_min=p_min)
            self._two_phase_limits = {
                "T": np.array([state.T for state in _two_phase_limits]),
                "h": np.array([state.h for state in _two_phase_limits]),
                "s": np.array([state.s for state in _two_phase_limits]),
                "p": np.array([state.p for state in _two_phase_limits]),
            }

        if quantity not in self._two_phase_limits:
            raise ValueError("The given quantity is not supported. T, h, s, or p are supported.")
        return self._two_phase_limits[quantity]

    def calc_mean_transport_properties(self, state_in, state_out):
        """
        Calculate the average transport properties for the given states.

        Args:
            state_in (ThermodynamicState): First state
            state_out (ThermodynamicState): Second state

        Returns:
            TransportProperties: Average transport properties

        Notes:
            The TransportProperties does not contain a state, as an average
            state is not possible to calculate.
        """
        tr_pr_in = self.calc_transport_properties(state_in)
        tr_pr_out = self.calc_transport_properties(state_out)

        return TransportProperties(
            lam=0.5 * (tr_pr_in.lam + tr_pr_out.lam),
            dyn_vis=0.5 * (tr_pr_in.dyn_vis + tr_pr_out.dyn_vis),
            kin_vis=0.5 * (tr_pr_in.kin_vis + tr_pr_out.kin_vis),
            pr=0.5 * (tr_pr_in.Pr + tr_pr_out.Pr),
            cp=0.5 * (tr_pr_in.cp + tr_pr_out.cp),
            cv=0.5 * (tr_pr_in.cv + tr_pr_out.cv),
            state=None)


def get_two_phase_limits(med_prop: MedProp, p_step: int = 1000, p_min: int = int(1e3)) -> List[ThermodynamicState]:
    """
    Return the states representing the boundaries of the two-phase section for the given fluid.

    This function is primarily used for visualizing the two-phase section and validating the accuracy of calculations.

    Args:
        med_prop (MedProp): An instance of a valid MedProp-Class.
        p_step (int): The step size for pressure variation in Pa. Default is 1000 Pa.
        p_min (int): The minimum pressure in Pa from where to start calculation. Default is 1000 Pa.

    Returns:
        List[ThermodynamicState]: A list of ThermodynamicState instances representing the two-phase limits.

    Notes:
        The two-phase limits are computed by iterating over a range of pressures from the minimum pressure up to the
        critical point pressure (exclusive) with a specified step size. States at quality 0 (saturated liquid)
        and quality 1 (saturated vapor) are appended to form the two-phase boundary. The list is reversed to
        maintain the correct order for visualization purposes.
    """
    _, _p_max, _ = med_prop.get_critical_point()
    q0, q1 = [], []
    for _p in range(p_min, int(_p_max), p_step):
        try:
            q0.append(med_prop.calc_state("PQ", _p, 0))
            q1.append(med_prop.calc_state("PQ", _p, 1))
        except ValueError as err:
            logger.info("Could not calculate two-phase limits for p=%s: %s",
                        _p, err)
    # Reverse list for correct order
    return q0 + q1[::-1]



"""
Module containing classes for thermodynamic state and transport properties.
"""
from vclibpy.datamodels import VariableContainer


__all__ = [
    'ThermodynamicState',
    'TransportProperties',
]


class ThermodynamicState:
    """
    Represents a thermodynamic state within a cycle.

    Notes:
        Does not necessarily need to have all state variables defined!

    Args:
        p (float): Pressure at the state in Pa.
        T (float): Temperature at the state in K.
        u (float): Inner energy at the state in J/kg.
        h (float): Enthalpy at the state in J/kg.
        s (float): Entropy at the state in J/(kg * K).
        v (float): Specific volume at the state in m^3/kg.
        q (float): Quality at the state (between 0 and 1).
        d (float): Density at the state in kg/m^3.

    Methods:
        __init__: Initializes the state class.
        __str__: Provides a string representation of the state.
        get_pretty_print: Formats the state with names, units, and descriptions.
    """

    def __init__(self,
                 p=None,
                 T=None,
                 u=None,
                 h=None,
                 s=None,
                 v=None,
                 q=None,
                 d=None):
        """
        Initializes a thermodynamic state.

        Args:
            p (float): Pressure at the state in Pa.
            T (float): Temperature at the state in K.
            u (float): Inner energy at the state in J/kg.
            h (float): Enthalpy at the state in J/kg.
            s (float): Entropy at the state in J/(kg * K).
            v (float): Specific volume at the state in m^3/kg.
            q (float): Quality at the state (between 0 and 1).
            d (float): Density at the state in kg/m^3.

        Notes:
            If only v or d is provided, the other attribute will be calculated. If both are given and they are similar,
            an error will be raised.
        """
        self.p = p
        self.T = T
        self.u = u
        self.h = h
        self.s = s
        self.v = v
        self.q = q
        self.d = d
        # Define density
        if v and d:
            if not round(1/v, 4) == round(d, 4):
                raise ValueError("At current state d and v do not match", d, v)
        elif v:
            self.d = 1/v
        elif d:
            self.v = 1/d

    def __str__(self):
        """
        Returns a string representation of the state.
        """
        return ";".join([f"{k}={v}" for k, v in self.__dict__.items()])

    def get_pretty_print(self):
        """
        Provides a formatted representation of the state with names, units, and descriptions.
        """
        _container = VariableContainer()
        _container.__class__.__name__ = self.__class__.__name__
        _container.set(name="p", value=self.p, unit="Pa", description="Pressure")
        _container.set(name="T", value=self.T, unit="K", description="Temperature")
        _container.set(name="u", value=self.u, unit="J/kg", description="Inner energy")
        _container.set(name="h", value=self.h, unit="J/kg", description="Enthalpy")
        _container.set(name="s", value=self.s, unit="J/(kg*K)", description="Entropy")
        _container.set(name="v", value=self.v, unit="m^3/kg", description="Specific volume")
        _container.set(name="q", value=self.q, unit="-", description="Quality")
        _container.set(name="d", value=self.d, unit="kg/m^3", description="Density")
        return str(_container)


class TransportProperties:
    """
    Represents transport properties at a specific thermodynamic state.

    Args:
        lam (float): Thermal conductivity in W/(m*K).
        dyn_vis (float): Dynamic viscosity in Pa*s.
        kin_vis (float): Kinematic viscosity in m^2/s.
        Pr (float): Prandtl number.
        cp (float): Isobaric specific heat capacity in J/(kg*K).
        cv (float): Isochoric specific heat capacity in J/(kg*K).
        beta (float): Thermal expansion coefficient in 1/K.
        sur_ten (float): Surface tension in N/m.
        ace_fac (float): Acentric factor.
        state (ThermodynamicState): The state the transport properties belong to.

    Methods:
        __init__: Initializes the transport properties class.
        __str__: Provides a string representation of the transport properties.
        get_pretty_print: Formats the properties with names, units, and descriptions.
    """

    def __init__(self,
                 lam=None,
                 dyn_vis=None,
                 kin_vis=None,
                 pr=None,
                 cp=None,
                 cv=None,
                 beta=None,
                 sur_ten=None,
                 ace_fac=None,
                 state=None):
        """
        Initializes transport properties.

        Args:
            lam (float): Thermal conductivity in W/(m*K).
            dyn_vis (float): Dynamic viscosity in Pa*s.
            kin_vis (float): Kinematic viscosity in m^2/s.
            pr (float): Prandtl number.
            cp (float): Isobaric specific heat capacity in J/(kg*K).
            cv (float): Isochoric specific heat capacity in J/(kg*K).
            beta (float): Thermal expansion coefficient in 1/K.
            sur_ten (float): Surface tension in N/m.
            ace_fac (float): Acentric factor.
            state (ThermodynamicState): The state the transport properties belong to.
        """
        self.lam = lam
        self.dyn_vis = dyn_vis
        self.kin_vis = kin_vis
        self.Pr = pr
        self.cp = cp
        self.cv = cv
        self.beta = beta
        self.sur_ten = sur_ten
        self.ace_fac = ace_fac
        self.state = state

    def __str__(self):
        """
        Returns a string representation of the transport properties.
        """
        return ";".join([f"{k}={v}" for k, v in self.__dict__.items()])

    def get_pretty_print(self):
        """
        Provides a formatted representation of the properties with names, units, and descriptions.
        """
        _container = VariableContainer()
        _container.__class__.__name__ = self.__class__.__name__
        _container.set(name="lam", value=self.lam, unit="W/(m*K)", description="Thermal conductivity")
        _container.set(name="dyn_vis", value=self.dyn_vis, unit="Pa*s", description="Dynamic viscosity")
        _container.set(name="kin_vis", value=self.kin_vis, unit="m^2/s", description="Kinematic viscosity")
        _container.set(name="pr", value=self.Pr, unit="-", description="Prandtl number")
        _container.set(name="cp", value=self.cp, unit="J/(kg*K)", description="Isobaric specific heat capacity")
        _container.set(name="cv", value=self.cv, unit="J/(kg*K)", description="Isochoric specific heat capacity")
        _container.set(name="beta", value=self.beta, unit="1/K", description="Thermal expansion coefficient")
        _container.set(name="sur_ten", value=self.sur_ten, unit="N/m", description="Surface tension")
        _container.set(name="ace_fac", value=self.ace_fac, unit="-", description="Acentric factor")
        return str(_container)



import numpy as np

from vclibpy.flowsheets.vapor_injection import BaseVaporInjection
from vclibpy.components.heat_exchangers.economizer import VaporInjectionEconomizerNTU


class VaporInjectionEconomizer(BaseVaporInjection):
    """
    Cycle with vapor injection using an economizer.

    For this cycle, we have 9 relevant states:

    - 1: Before compressor, after evaporator
    - 2: Before condenser, after compressor
    - 3: Before ihx, after condenser
    - 4: Before Evaporator, after ihx
    - 5_ihx: Before ihx, after condenser and EV
    - 6_ihx: Before Mixing with 1_VI, after ihx
    - 7_ihx: Before second EV, after ihx
    - 1_VI: Before Mixing with 6_ihx, After first stage
    - 1_VI_mixed. Before second stage of compressor, after mixing with 6_ihx

    Additional Assumptions:
    -----------------------
    - No heat losses in ihx
    - No pressure loss in ihx
    - No losses for splitting of streams
    - Isenthalpic second EV

    Notes
    -----
    See parent docstring for info on further assumptions and parameters.
    """

    flowsheet_name = "VaporInjectionEconomizer"

    def __init__(self, economizer: VaporInjectionEconomizerNTU, **kwargs):
        self.economizer = economizer
        super().__init__(**kwargs)

    def get_all_components(self):
        return super().get_all_components() + [
            self.economizer
        ]

    def calc_injection(self):
        """
        This calculation assumes that the heat transfer
        of the higher temperature liquid is always in the subcooling
        region, while the vapor injection is always a two-phase heat
        transfer.
        In reality, you would need to achieve a certain degree of superheat.
        Thus, a moving boundary approach would be more fitting. For simplicity,
        we assume no superheat.

        This function iterates the amount of vapor injected.
        The iteration starts with close to no injection and increases
        the amount of injection as long as enough hotter sub-cooled liquid is
        present to fully vaporize the injected part.
        """
        self.economizer.state_inlet = self.condenser.state_outlet
        self.economizer.state_two_phase_inlet = self.high_pressure_valve.state_outlet
        self.economizer.state_two_phase_outlet = self.med_prop.calc_state(
            "PQ", self.high_pressure_valve.state_outlet.p, 1
        )
        m_flow_evaporator = self.evaporator.m_flow

        dh_ihe_goal = (
                self.economizer.state_two_phase_outlet.h -
                self.economizer.state_two_phase_inlet.h
        )

        # Get transport properties:
        tra_properties_liquid = self.med_prop.calc_transport_properties(
            self.economizer.state_inlet
        )
        alpha_liquid = self.economizer.calc_alpha_liquid(tra_properties_liquid)
        tra_properties_two_phase = self.med_prop.calc_mean_transport_properties(
            self.economizer.state_two_phase_inlet,
            self.economizer.state_two_phase_outlet
        )
        alpha_two_phase = self.economizer.calc_alpha_liquid(tra_properties_two_phase)

        # Set cp based on transport properties
        dT_secondary = (
                self.economizer.state_two_phase_outlet.T -
                self.economizer.state_two_phase_inlet.T
        )
        if dT_secondary == 0:
            cp_4 = np.inf
        else:
            cp_4 = dh_ihe_goal / dT_secondary
        self.economizer.set_secondary_cp(cp=cp_4)
        self.economizer.set_primary_cp(cp=tra_properties_liquid.cp)

        # We have to iterate to ensure the correct fraction of mass is
        # used to ensure state5 has q=1
        _x_vi_step = 0.1
        _min_step_x_vi = 0.0001

        x_vi_next = _min_step_x_vi  # Don't start with zero!
        while True:
            x_vi = x_vi_next
            x_eva = 1 - x_vi
            m_flow_vapor_injection = x_vi * m_flow_evaporator
            Q_flow_goal = dh_ihe_goal * m_flow_vapor_injection

            self.economizer.m_flow = x_eva * m_flow_evaporator
            self.economizer.m_flow_secondary = m_flow_vapor_injection

            # This dT_max is always valid, as the primary inlet is cooled
            # and the secondary inlet (the vapor) is either heated
            # or isothermal for pure fluids
            Q_flow, k = self.economizer.calc_Q_ntu(
                dT_max=(
                        self.economizer.state_inlet.T -
                        self.economizer.state_two_phase_inlet.T
                ),
                alpha_pri=alpha_liquid,
                alpha_sec=alpha_two_phase,
                A=self.economizer.A
            )
            if Q_flow > Q_flow_goal:
                if _x_vi_step <= _min_step_x_vi:
                    break
                # We can increase x_vi_next further, as more heat can be extracted
                x_vi_next = x_vi + _x_vi_step
            else:
                x_vi_next = x_vi - _x_vi_step * 0.9
                _x_vi_step /= 10

        # Solve Energy Balance
        h_7 = self.economizer.state_inlet.h - dh_ihe_goal * self.economizer.m_flow
        state_7_ihx = self.med_prop.calc_state("PH", self.economizer.state_inlet.p, h_7)
        self.economizer.state_outlet = state_7_ihx
        return x_vi, self.economizer.state_two_phase_outlet.h, state_7_ihx

    def get_states_in_order_for_plotting(self):
        return super().get_states_in_order_for_plotting() + [
            self.economizer.state_two_phase_inlet,
            self.economizer.state_two_phase_outlet,
            self.high_pressure_compressor.state_inlet,
            # Go back to the condenser outlet
            self.economizer.state_two_phase_outlet,
            self.economizer.state_two_phase_inlet,
            self.high_pressure_valve.state_outlet,
            self.high_pressure_valve.state_inlet,
            self.condenser.state_outlet,
            self.economizer.state_inlet,
            self.economizer.state_outlet
        ]



import shutil
import unittest
import logging
import pathlib
import os
from random import random

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from vclibpy import utils
from vclibpy import datamodels
from vclibpy.components.expansion_valves import Bernoulli
from vclibpy.components.compressors import RotaryCompressor
from vclibpy.components.heat_exchangers.economizer import VaporInjectionEconomizerNTU
from vclibpy.flowsheets import StandardCycle, VaporInjectionEconomizer, VaporInjectionPhaseSeparator
from vclibpy.components.heat_exchangers import moving_boundary_ntu
from vclibpy.components.heat_exchangers import heat_transfer


def _load_flowsheet(fluid: str, flowsheet: str = None):
    flowsheets = {
        "Standard": StandardCycle,
        "VaporInjectionPhaseSeparator": VaporInjectionPhaseSeparator,
        "VaporInjectionEconomizer": VaporInjectionEconomizer
    }
    N_max = 125
    V_h = 19e-6
    compressor = RotaryCompressor(
        N_max=N_max,
        V_h=V_h
    )
    condenser = moving_boundary_ntu.MovingBoundaryNTUCondenser(
        A=5,
        secondary_medium="water",
        flow_type="counter",
        ratio_outer_to_inner_area=1,
        two_phase_heat_transfer=heat_transfer.constant.ConstantTwoPhaseHeatTransfer(alpha=5000),
        gas_heat_transfer=heat_transfer.constant.ConstantHeatTransfer(alpha=5000),
        wall_heat_transfer=heat_transfer.wall.WallTransfer(lambda_=236, thickness=2e-3),
        liquid_heat_transfer=heat_transfer.constant.ConstantHeatTransfer(alpha=5000),
        secondary_heat_transfer=heat_transfer.constant.ConstantHeatTransfer(alpha=5000)
    )
    evaporator = moving_boundary_ntu.MovingBoundaryNTUEvaporator(
        A=15,
        secondary_medium="air",
        flow_type="counter",
        ratio_outer_to_inner_area=1,
        two_phase_heat_transfer=heat_transfer.constant.ConstantTwoPhaseHeatTransfer(alpha=5000),
        gas_heat_transfer=heat_transfer.constant.ConstantHeatTransfer(alpha=5000),
        wall_heat_transfer=heat_transfer.wall.WallTransfer(lambda_=236, thickness=2e-3),
        liquid_heat_transfer=heat_transfer.constant.ConstantHeatTransfer(alpha=5000),
        secondary_heat_transfer=heat_transfer.constant.ConstantHeatTransfer(alpha=5000)
    )
    economizer = VaporInjectionEconomizerNTU(
        A=2,
        two_phase_heat_transfer=heat_transfer.constant.ConstantTwoPhaseHeatTransfer(alpha=50000),
        gas_heat_transfer=heat_transfer.constant.ConstantHeatTransfer(alpha=50000),
        wall_heat_transfer=heat_transfer.wall.WallTransfer(lambda_=236, thickness=2e-3),
        liquid_heat_transfer=heat_transfer.constant.ConstantHeatTransfer(alpha=50000),
    )
    expansion_valve = Bernoulli(A=0.1)  # Should not matter
    kwargs = dict(
        evaporator=evaporator,
        condenser=condenser,
        fluid=fluid
    )
    if flowsheet == "Standard":
        kwargs.update(dict(
            expansion_valve=expansion_valve,
            compressor=compressor
        ))
    else:
        kwargs.update(dict(
            high_pressure_compressor=compressor,
            low_pressure_compressor=compressor,
            high_pressure_valve=expansion_valve,
            low_pressure_valve=expansion_valve
        ))
    if flowsheet == "VaporInjectionEconomizer":
        kwargs.update(dict(
            economizer=economizer
        ))
    return flowsheets[flowsheet](**kwargs)


class TestRegressionWithAllFluidsAndFlowsheets(unittest.TestCase):
    """
    Return the settings used in the master thesis and following
    publication: https://doi.org/10.1016/j.enconman.2021.114888
    """

    def setUp(self) -> None:
        self.fs_state = datamodels.FlowsheetState()
        self.var = datamodels.Variable(name="test", value=random(), unit="K", description="dummy")
        import uuid
        self.working_dir = pathlib.Path(__file__).parent.joinpath(str(uuid.uuid4()))
        os.makedirs(self.working_dir, exist_ok=True)

        self.inputs_to_compare = [
            "n in - (Relative compressor speed)",
            "T_eva_in in K (Secondary side evaporator inlet temperature)",
            "T_con_in in K (Secondary side condenser inlet temperature)",
        ]

    def _regression_of_examples(self, flowsheet, fluid):
        self.results_to_compare = [
            "Q_con in W (Condenser heat flow rate)",
            "COP in - (Coefficient of performance)",
            "m_flow_ref in kg/s (Refrigerant mass flow rate)",
            "T_1 in K (Refrigerant temperature at evaporator outlet)",
            "T_2 in K (Compressor outlet temperature)",
            "T_3 in K (Refrigerant temperature at condenser outlet)",
            "T_4 in K (Refrigerant temperature at evaporator inlet)",
            "p_con in Pa (Condensation pressure)",
            "p_eva in Pa (Evaporation pressure)",
            "A_eva_sh in m2 (Area for superheat heat exchange in evaporator)",
            "A_eva_lat in m2 (Area for latent heat exchange in evaporator)",
            "A_con_sh in m2 (Area for superheat heat exchange in condenser)",
            "A_con_lat in m2 (Area for latent heat exchange in condenser)",
            "A_con_sc in m2 (Area for subcooling heat exchange in condenser)",
            "carnot_quality in - (Carnot Quality)",
        ]

        # Select the settings / parameters of the hp:
        kwargs = {"max_err_ntu": 0.5,
                  "max_err_dT_min": 0.1,
                  "show_iteration": False,
                  "max_num_iterations": 5000}

        # Just for quick study: Specify concrete points:
        T_eva_in_ar = [-10 + 273.15, 273.15]
        T_con_in_ar = [30 + 273.15, 70 + 273.15]
        n_ar = [0.3, 1]

        os.makedirs(self.working_dir, exist_ok=True)

        kwargs["save_path_plots"] = pathlib.Path(self.working_dir).joinpath("plots")
        os.makedirs(pathlib.Path(self.working_dir).joinpath("plots"), exist_ok=True)

        heat_pump = _load_flowsheet(
            fluid=fluid,
            flowsheet=flowsheet
        )
        _, path_csv = utils.full_factorial_map_generation(
            heat_pump=heat_pump,
            save_path=self.working_dir,
            T_con_in_ar=T_con_in_ar,
            T_eva_in_ar=T_eva_in_ar,
            n_ar=n_ar,
            use_multiprocessing=False,
            save_plots=False,
            m_flow_con=0.2,
            m_flow_eva=0.9,
            dT_eva_superheating=5,
            dT_con_subcooling=0,
        )
        path_csv_regression = pathlib.Path(__file__).parent.joinpath(
            "regression_data", "reference_results", f"{flowsheet}_{fluid}.csv"
        )
        self.assertTrue(self._compare_results(path_csv, path_csv_regression))

    def test_standard_propane(self):
        self._regression_of_examples("Standard", "Propane")

    def test_standard_R410A(self):
        self._regression_of_examples("Standard", "R410A")

    def test_vi_ps_propane(self):
        self._regression_of_examples("VaporInjectionPhaseSeparator", "Propane")

    def test_evi_propane(self):
        self.skipTest("EVI works locally, only CI fails.")
        self._regression_of_examples("VaporInjectionEconomizer", "Propane")

    @unittest.skip("not implemented")
    def test_opti_horst_regression(self):
        self.results_to_compare = [
            "Q_con in W (Condenser heat flow rate)",
            "COP in - (Coefficient of performance)",
            "m_flow_ref in kg/s (Refrigerant mass flow rate)",
            "T_1 in K (Refrigerant temperature at compressor inlet)",
            "T_2 in K (Compressor outlet temperature)",
            "T_3 in K (Refrigerant temperature at condenser outlet)",
            "T_4 in K (Refrigerant temperature at evaporator inlet)",
            "p_con in Pa (Condensation pressure)",
            "p_eva in Pa (Evaporation pressure)",
            "p_2 in Pa (Pressure after compressor)",
            "A_eva_sh in m2 (Area for superheat heat exchange in evaporator)",
            "A_eva_lat in m2 (Area for latent heat exchange in evaporator)",
            "A_con_sh in m2 (Area for superheat heat exchange in condenser)",
            "A_con_lat in m2 (Area for latent heat exchange in condenser)",
            "A_con_sc in m2 (Area for subcooling heat exchange in condenser)",
            "carnot_quality in - (Carnot Quality)",
        ]
        path_csv = pathlib.Path(__file__).parent.joinpath(
            "regression_data", "temp", "OptiHorst_R410A.csv"
        )
        path_csv_regression = pathlib.Path(__file__).parent.joinpath(
            "regression_data", "publication", "OptiHorst_R410A.csv"
        )
        no_error = self._compare_results(path_csv, path_csv_regression)
        self.assertTrue(no_error)

    def _compare_results(self, path_csv, path_csv_regression):
        df = pd.read_csv(path_csv, index_col=0)
        df_regression = pd.read_csv(path_csv_regression, index_col=0)
        # Compare if all inputs are present
        # Use this result object to pretty print any errors
        results_comparison = {col: {} for col in self.results_to_compare}
        for idx, row in df.iterrows():
            input_mask = np.ones(len(df_regression)) == 1  # All True
            for inp in self.inputs_to_compare:
                # We use is_close as floating errors may lead to 0.6 being 0.600000000002
                is_close = np.isclose(df_regression.loc[:, inp], row[inp])
                self.assertTrue(np.any(is_close))
                input_mask = input_mask & is_close
            for col in self.results_to_compare:
                new_value = row[col]
                old_value = df_regression.loc[input_mask, col].values[0]
                if np.isnan(new_value) and old_value in [np.nan, 1, 0]:
                    continue
                results_comparison[col][new_value] = old_value
        no_errors = []
        error_info = []
        for col in self.results_to_compare:
            res_new = np.array(list(results_comparison[col].keys()))
            res_old = np.array(list(results_comparison[col].values()))
            nan_mask = np.isnan(res_old)
            res_new = res_new[~nan_mask]
            res_old = res_old[~nan_mask]
            zero_mask = res_old == 0
            res_new = res_new[~zero_mask]
            res_old = res_old[~zero_mask]

            # Smaller 1 percent error
            deviation = (res_new - res_old) / res_old * 100
            deviation_ok = deviation < 1
            no_regression_error = np.all(deviation_ok)
            if not no_regression_error:
                error_info.append(f"{col}: {deviation[~deviation_ok]}")
                plt.figure()
                plt.plot(res_old[~deviation_ok], label="old results", marker="s")
                plt.plot(res_new[~deviation_ok], label="New results", marker="o")
                plt.ylabel(col)
                plt.legend()
                os.makedirs(self.working_dir.parent.joinpath("regression_errors"), exist_ok=True)
                col_clean = col.replace("/", "_")
                plt.savefig(self.working_dir.parent.joinpath("regression_errors", f"{path_csv.stem}_{col_clean}.png"))
                plt.close(plt.gcf())
            no_errors.append(no_regression_error)
        if not np.all(no_errors):
            print("Errors for the following variables:")
            for _error_info in error_info:
                print(_error_info)
        return np.all(no_errors)


def tearDown(self) -> None:
    try:
        shutil.rmtree(self.working_dir)
    except PermissionError:
        logging.error(
            "Could not delete folders due to PermissionError, delete them yourself: %s",
            self.working_dir
        )


if __name__ == "__main__":
    unittest.main()



"""Module with functions to test the if examples are
executable."""
import importlib
import sys
import unittest
import subprocess
import pathlib


class TestExamples(unittest.TestCase):
    """Test-class for the examples of vclibpy."""

    def setUp(self) -> None:
        self.timeout = 100  # Seconds which the script is allowed to run

    def _run_example(self, example, timeout=None):
        ex_py = pathlib.Path(__file__).absolute().parents[1].joinpath("examples", example)
        sys.path.insert(0, str(ex_py.parent))
        module = importlib.import_module(ex_py.stem)
        example_main = getattr(module, "main")
        example_main()

    def test_e1_refrigerant_data(self):
        self._run_example(example="e1_refrigerant_data.py")

    def test_e2_compressor(self):
        self._run_example(example="e2_compressor.py")

    def test_e3_inputs_and_flowsheet_state(self):
        self._run_example(example="e3_inputs_and_flowsheet_state.py")

    def test_e4_heat_exchanger(self):
        self._run_example(example="e4_heat_exchanger.py")

    def test_e5_expansion_valve(self):
        self._run_example(example="e5_expansion_valve.py")

    def test_e6_simple_heat_pump(self):
        self._run_example(example="e6_simple_heat_pump.py")

    def test_e7_vapor_injection(self):
        self._run_example(example="e7_vapor_injection.py")


if __name__ == "__main__":
    unittest.main()



import logging

from vclibpy.flowsheets.vapor_injection import BaseVaporInjection
from vclibpy.components.phase_separator import PhaseSeparator


logger = logging.getLogger(__name__)


class VaporInjectionPhaseSeparator(BaseVaporInjection):
    """
    Cycle with vapor injection using an adiabatic ideal phase seperator.

    For this cycle, we have 9 relevant states:

    - 1: Before compressor, after evaporator
    - 2: Before condenser, after compressor
    - 3: Before PS, after condenser
    - 4: Before Evaporator, after PS
    - 5_vips: Before PS, after first EV
    - 6_vips: Before Mixing with 1_VI, after PS
    - 7_vips: Before second EV, after PS
    - 1_VI: Before Mixing with 6_vips, After first stage
    - 1_VI_mixed. Before second stage of compressor, after mixing with 6_vips

    Additional Assumptions:
    -----------------------
    - Ideal mixing in compressor of state 5 and state 4

    Notes
    -----
    See parent docstring for info on further assumptions and parameters.
    """

    flowsheet_name = "VaporInjectionPhaseSeparator"

    def __init__(self, **kwargs):
        self.phase_separator = PhaseSeparator()
        super().__init__(**kwargs)

    def get_all_components(self):
        return super().get_all_components() + [
            self.phase_separator
        ]

    def calc_injection(self):
        # Phase separator
        self.phase_separator.state_inlet = self.high_pressure_valve.state_outlet
        x_vapor_injection = self.phase_separator.state_inlet.q
        h_vapor_injection = self.phase_separator.state_outlet_vapor.h
        return x_vapor_injection, h_vapor_injection, self.phase_separator.state_outlet_liquid

    def get_states_in_order_for_plotting(self):
        return super().get_states_in_order_for_plotting() + [
            self.phase_separator.state_inlet,
            self.phase_separator.state_outlet_vapor,
            self.high_pressure_compressor.state_inlet,
            # Go back to separator for clear lines
            self.phase_separator.state_outlet_vapor,
            self.phase_separator.state_outlet_liquid
        ]



"""
Module for different compressor models
"""

from vclibpy.components.component import BaseComponent
from vclibpy.datamodels import Inputs, FlowsheetState


class Compressor(BaseComponent):
    """
    Base compressor class to be extended for specific compressor models.

    Args:
        N_max (float): Maximal rotations per second of the compressor.
        V_h (float): Volume of the compressor in m^3.

    Methods:
        get_lambda_h(inputs: Inputs) -> float:
            Get the volumetric efficiency.

        get_eta_isentropic(p_outlet: float, inputs: Inputs) -> float:
            Get the isentropic efficiency.

        get_eta_mech(inputs: Inputs) -> float:
            Get the mechanical efficiency.

        get_p_outlet() -> float:
            Get the outlet pressure.

        get_n_absolute(n: float) -> float:
            Return the absolute compressor frequency based on the relative speed.

        calc_state_outlet(p_outlet: float, inputs: Inputs, fs_state: FlowsheetState):
            Calculate the outlet state based on the high pressure level and provided inputs.

        calc_m_flow(inputs: Inputs, fs_state: FlowsheetState) -> float:
            Calculate the refrigerant mass flow rate.

        calc_electrical_power(inputs: Inputs, fs_state: FlowsheetState) -> float:
            Calculate the electrical power consumed by the compressor based on an adiabatic energy balance.
    """

    def __init__(self, N_max: float, V_h: float):
        """
        Initialize the compressor.

        Args:
            N_max (float): Maximal rotations per second of the compressor.
            V_h (float): Volume of the compressor in m^3.
        """
        super().__init__()
        self.N_max = N_max
        self.V_h = V_h

    def get_lambda_h(self, inputs: Inputs) -> float:
        """
        Get the volumetric efficiency.

        Args:
            inputs (Inputs): Inputs for the calculation.

        Returns:
            float: Volumetric efficiency.
        """
        raise NotImplementedError("Re-implement this function to use it")

    def get_eta_isentropic(self, p_outlet: float, inputs: Inputs) -> float:
        """
        Get the isentropic efficiency.

        Args:
            p_outlet (float): High pressure value.
            inputs (Inputs): Inputs for the calculation.

        Returns:
            float: Isentropic efficiency.
        """
        raise NotImplementedError("Re-implement this function to use it")

    def get_eta_mech(self, inputs: Inputs) -> float:
        """
        Get the mechanical efficiency including motor and inverter efficiencies.

        Args:
            inputs (Inputs): Inputs for the calculation.

        Returns:
            float: Mechanical efficiency including motor and inverter efficiencies.
        """
        raise NotImplementedError("Re-implement this function to use it")

    def get_p_outlet(self) -> float:
        """
        Get the outlet pressure.

        Returns:
            float: Outlet pressure.
        """
        assert self.state_outlet is not None, "You have to calculate the outlet state first."
        return self.state_outlet.p

    def get_n_absolute(self, n: float) -> float:
        """
        Return given relative n as absolute rounds/sec based on self.N_max.

        Args:
            n (float): Relative compressor speed between 0 and 1.

        Returns:
            float: Absolute compressor frequency in rounds/sec.
        """
        return self.N_max * n

    def calc_state_outlet(self, p_outlet: float, inputs: Inputs, fs_state: FlowsheetState):
        """
        Calculate the output state based on the high pressure level and the provided inputs.
        The state is automatically set as the outlet state of this component.

        Args:
            p_outlet (float): High pressure value.
            inputs (Inputs): Inputs for calculation.
            fs_state (FlowsheetState): Flowsheet state.
        """
        state_outlet_isentropic = self.med_prop.calc_state("PS", p_outlet, self.state_inlet.s)
        eta_is = self.get_eta_isentropic(p_outlet=p_outlet, inputs=inputs)
        h_outlet = (
                self.state_inlet.h + (state_outlet_isentropic.h - self.state_inlet.h) /
                eta_is
        )
        fs_state.set(name="eta_is", value=eta_is, unit="%", description="Isentropic efficiency")
        self.state_outlet = self.med_prop.calc_state("PH", p_outlet, h_outlet)

    def calc_m_flow(self, inputs: Inputs, fs_state: FlowsheetState) -> float:
        """
        Calculate the refrigerant mass flow rate.

        Args:
            inputs (Inputs): Inputs for the calculation.
            fs_state (FlowsheetState): Flowsheet state.

        Returns:
            float: Refrigerant mass flow rate.
        """
        lambda_h = self.get_lambda_h(inputs=inputs)
        V_flow_ref = (
                lambda_h *
                self.V_h *
                self.get_n_absolute(inputs.n)
        )
        self.m_flow = self.state_inlet.d * V_flow_ref
        fs_state.set(name="lambda_h", value=lambda_h, unit="%", description="Volumetric efficiency")
        fs_state.set(name="V_flow_ref", value=V_flow_ref, unit="m3/s", description="Refrigerant volume flow rate")
        fs_state.set(name="m_flow_ref", value=self.m_flow, unit="kg/s", description="Refrigerant mass flow rate")
        return self.m_flow

    def calc_electrical_power(self, inputs: Inputs, fs_state: FlowsheetState) -> float:
        """
        Calculate the electrical power consumed by the compressor based on an adiabatic energy balance.

        Args:
            inputs (Inputs): Inputs for the calculation.
            fs_state (FlowsheetState): Flowsheet state.

        Returns:
            float: Electrical power consumed.
        """
        # Heat flow in the compressor
        P_t = self.m_flow * (self.state_outlet.h - self.state_inlet.h)
        # Electrical power consumed
        eta_mech = self.get_eta_mech(inputs=inputs)
        P_el = P_t / eta_mech
        fs_state.set(name="eta_mech", value=eta_mech, unit="-", description="Mechanical efficiency")
        return P_el



from vclibpy.flowsheets import BaseCycle
from vclibpy.datamodels import FlowsheetState, Inputs
from vclibpy.components.compressors import Compressor
from vclibpy.components.expansion_valves import ExpansionValve


class StandardCycle(BaseCycle):
    """
    Class for a standard cycle with four components.

    For the standard cycle, we have 4 possible states:

    1. Before compressor, after evaporator
    2. Before condenser, after compressor
    3. Before EV, after condenser
    4. Before Evaporator, after EV
    """

    flowsheet_name = "Standard"

    def __init__(
            self,
            compressor: Compressor,
            expansion_valve: ExpansionValve,
            **kwargs
    ):
        super().__init__(**kwargs)
        self.compressor = compressor
        self.expansion_valve = expansion_valve

    def get_all_components(self):
        return super().get_all_components() + [
            self.compressor,
            self.expansion_valve
        ]

    def get_states_in_order_for_plotting(self):
        return [
            self.evaporator.state_inlet,
            self.med_prop.calc_state("PQ", self.evaporator.state_inlet.p, 1),
            self.evaporator.state_outlet,
            self.compressor.state_inlet,
            self.compressor.state_outlet,
            self.condenser.state_inlet,
            self.med_prop.calc_state("PQ", self.condenser.state_inlet.p, 1),
            self.med_prop.calc_state("PQ", self.condenser.state_inlet.p, 0),
            self.condenser.state_outlet,
            self.expansion_valve.state_inlet,
            self.expansion_valve.state_outlet,
        ]

    def calc_states(self, p_1, p_2, inputs: Inputs, fs_state: FlowsheetState):
        self.set_condenser_outlet_based_on_subcooling(p_con=p_2, inputs=inputs)
        self.expansion_valve.state_inlet = self.condenser.state_outlet
        self.expansion_valve.calc_outlet(p_outlet=p_1)
        self.evaporator.state_inlet = self.expansion_valve.state_outlet
        self.set_evaporator_outlet_based_on_superheating(p_eva=p_1, inputs=inputs)
        self.compressor.state_inlet = self.evaporator.state_outlet
        self.compressor.calc_state_outlet(p_outlet=p_2, inputs=inputs, fs_state=fs_state)
        self.condenser.state_inlet = self.compressor.state_outlet

        # Mass flow rate:
        self.compressor.calc_m_flow(inputs=inputs, fs_state=fs_state)
        self.condenser.m_flow = self.compressor.m_flow
        self.evaporator.m_flow = self.compressor.m_flow
        self.expansion_valve.m_flow = self.compressor.m_flow
        fs_state.set(
            name="y_EV", value=self.expansion_valve.calc_opening_at_m_flow(m_flow=self.expansion_valve.m_flow),
            unit="-", description="Expansion valve opening"
        )
        fs_state.set(
            name="T_1", value=self.evaporator.state_outlet.T,
            unit="K", description="Refrigerant temperature at evaporator outlet"
        )
        fs_state.set(
            name="T_2", value=self.compressor.state_outlet.T,
            unit="K", description="Compressor outlet temperature"
        )
        fs_state.set(
            name="T_3", value=self.condenser.state_outlet.T, unit="K",
            description="Refrigerant temperature at condenser outlet"
        )
        fs_state.set(
            name="T_4", value=self.evaporator.state_inlet.T,
            unit="K", description="Refrigerant temperature at evaporator inlet"
        )
        fs_state.set(name="p_con", value=p_2, unit="Pa", description="Condensation pressure")
        fs_state.set(name="p_eva", value=p_1, unit="Pa", description="Evaporation pressure")

    def calc_electrical_power(self, inputs: Inputs, fs_state: FlowsheetState):
        """Based on simple energy balance - Adiabatic"""
        return self.compressor.calc_electrical_power(inputs=inputs, fs_state=fs_state)



import time
import logging

from vclibpy import Inputs
from vclibpy.flowsheets import BaseCycle

logger = logging.getLogger(__name__)


def nominal_hp_design(
        heat_pump: BaseCycle, 
        inputs: Inputs, 
        fluid: str,
        dT_con: float = None,
        dT_eva: float = None,
        **kwargs) -> dict:
    """
    Function to calculate the heat pump design
    at a given nominal point. 
    Args:
        heat_pump (BaseCycle): A supported flowsheet
        inputs (Inputs):
            The input values at the nominal point.
            If the mass flow rates are not given, you
            can use dT_con and dT_eva to iteratively calculate
            the mass flow rates in order to achieve the required
            temperature differences at the nominal point.
        dT_con (float):
            Condenser temperature difference to calculate mass flow rate
        dT_eva (float):
            Evaporator temperature difference to calculate mass flow rate
        fluid (str): Fluid to be used.
        **kwargs: 
            m_flow_eva_start: Guess start-value for iteration. Default 0.2
            m_flow_con_start: Guess start-value for iteration. Default 1
            accuracy: Minimal accuracy for mass flow rate iteration (in kg/s). 
                Default 0.001 kg/s
             
    Returns:
        dict: A dictionary with all flowsheet states and inputs containing
              information about the nominal design.
    """
    t0 = time.time()
    # Define nominal values:
    m_flow_con_start = kwargs.get("m_flow_con_start", 0.2)
    m_flow_eva_start = kwargs.get("m_flow_eva_start", 1)
    accuracy = kwargs.get("accuracy", 0.001)
    calculate_m_flow = dT_eva is not None and dT_con is not None
    if calculate_m_flow:
        # Get nominal value:
        fs_state = heat_pump.calc_steady_state(fluid=fluid, inputs=inputs)
        if fs_state is None:
            raise ValueError("Given configuration is infeasible at nominal point.")

    else:
        # We have to iterate to match the m_flows to the Q_cons:
        m_flow_eva_next = m_flow_eva_start
        m_flow_con_next = m_flow_con_start
        while True:
            # Set values
            m_flow_eva = m_flow_eva_next
            m_flow_con = m_flow_con_next
            inputs.set("m_flow_con", m_flow_con)
            inputs.set("m_flow_eva", m_flow_eva)
            # Get nominal value:
            fs_state = heat_pump.calc_steady_state(fluid=fluid, inputs=inputs)
            if fs_state is None:
                raise ValueError("Given configuration is infeasible at nominal point.")
            cp_eva = heat_pump.evaporator._secondary_cp
            cp_con = heat_pump.condenser._secondary_cp
            m_flow_con_next = fs_state.get("Q_con") / (dT_con * cp_con)
            m_flow_eva_next = (fs_state.get("Q_con") * (1 - 1 / fs_state.get("COP"))) / (dT_eva * cp_eva)
            # Check convergence:
            if abs(m_flow_eva_next - m_flow_eva) < accuracy and abs(m_flow_con-m_flow_con_next) < accuracy:
                break

    nominal_design_info = {
        **inputs.convert_to_str_value_format(with_unit_and_description=False),
        **fs_state.convert_to_str_value_format(with_unit_and_description=False),
        "dT_con": dT_con,
        "dT_eva": dT_eva
    }
    logger.info("Auto-generation of nominal values took %s seconds", time.time()-t0)
    logger.info('Nominal values: %s', nominal_design_info)

    return nominal_design_info



import logging

from .heat_transfer import HeatTransfer
from vclibpy.media import TransportProperties


logger = logging.getLogger(__name__)


class WallTransfer(HeatTransfer):
    """
    Simple wall heat transfer

    Args:
        lambda_ (float):
            Heat conductivity of wall material in W/Km
        thickness (float):
            Thickness of wall in m^2
    """
    def __init__(self, lambda_: float, thickness: float):
        self.thickness = thickness
        self.lambda_ = lambda_

    def calc(self, transport_properties: TransportProperties, m_flow: float) -> float:
        """
        Heat transfer coefficient inside wall

        Returns:
            float: Heat transfer coefficient in W/(m^2*K)
        """
        return self.lambda_ / self.thickness



from .compressor import Compressor
from .rotary import RotaryCompressor
from .ten_coefficient import TenCoefficientCompressor, DataSheetCompressor
from .constant_effectivness import ConstantEffectivenessCompressor



from vclibpy.components.compressors.compressor import Compressor
from vclibpy.datamodels import Inputs


class ConstantEffectivenessCompressor(Compressor):
    """
    Compressor model with constant efficiencies.

    Inherits from the Compressor class, which defines the basic properties and behavior of a compressor in a vapor
    compression cycle.

    Parameters:
        N_max (float): Maximal rotations per second of the compressor.
        V_h (float): Volume of the compressor in m^3.
        eta_isentropic (float): Constant isentropic efficiency of the compressor.
        eta_mech (float): Constant mechanical efficiency of the compressor.
        lambda_h (float): Constant volumetric efficiency.

    Args:
        N_max (float): Maximal rotations per second of the compressor.
        V_h (float): Volume of the compressor in m^3.
        eta_isentropic (float): Constant isentropic efficiency of the compressor.
        eta_inverter (float): Constant inverter efficiency of the compressor.
        eta_motor (float): Constant motor efficiency of the compressor.
        eta_mech (float): Constant mechanical efficiency of the compressor including motor and inverter efficiencies.
        lambda_h (float): Constant volumetric efficiency.

    Methods:
        get_lambda_h(inputs: Inputs) -> float:
            Returns the constant volumetric efficiency of the compressor.

        get_eta_isentropic(p_outlet: float, inputs: Inputs) -> float:
            Returns the constant isentropic efficiency of the compressor.

        get_eta_mech(inputs: Inputs) -> float:
            Returns the constant mechanical efficiency including motor and inverter efficiencies.

    """

    def __init__(self,
                 N_max: float, V_h: float,
                 eta_isentropic: float,
                 eta_mech: float,
                 lambda_h: float):
        """
        Initialize the ConstantEffectivenessCompressor.

        Args:
            N_max (float): Maximal rotations per second of the compressor.
            V_h (float): Volume of the compressor in m^3.
            eta_isentropic (float): Constant isentropic efficiency of the compressor.
            eta_inverter (float): Constant inverter efficiency of the compressor.
            eta_motor (float): Constant motor efficiency of the compressor.
            eta_mech (float): Constant mechanical efficiency of the compressor.
            lambda_h (float): Constant volumetric efficiency.
        """
        super().__init__(N_max=N_max, V_h=V_h)
        self.eta_isentropic = eta_isentropic
        self.eta_mech = eta_mech
        self.lambda_h = lambda_h

    def get_lambda_h(self, inputs: Inputs) -> float:
        """
        Returns the constant volumetric efficiency of the compressor.

        Args:
            inputs (Inputs): Input parameters for the calculation.

        Returns:
            float: Constant volumetric efficiency.
        """
        return self.lambda_h

    def get_eta_isentropic(self, p_outlet: float, inputs: Inputs) -> float:
        """
        Returns the constant isentropic efficiency of the compressor.

        Args:
            p_outlet (float): High pressure value.
            inputs (Inputs): Input parameters for the calculation.

        Returns:
            float: Constant isentropic efficiency.
        """
        return self.eta_isentropic

    def get_eta_mech(self, inputs: Inputs) -> float:
        """
        Returns the product of the constant mechanical, motor, and inverter efficiencies
        as the effective mechanical efficiency of the compressor.

        Args:
            inputs (Inputs): Input parameters for the calculation.

        Returns:
            float: Effective mechanical efficiency.
        """
        return self.eta_mech



"""
Module which contains datamodels which are used in this library.
"""

from dataclasses import dataclass
from typing import Dict, Any
from copy import deepcopy


@dataclass
class Variable:
    """
    Class for a variable used in analysis.

    Args:
        name (str): The name of the variable.
        value (float): The numerical value of the variable.
        unit (str): The unit of measurement for the variable (optional).
        description (str): A description of the variable (optional).
    """
    name: str
    value: float
    unit: str = None
    description: str = None


class VariableContainer:
    """
    Class which holds Variables to be used anywhere in the models.

    This class enables dynamic addition of Variables.
    """
    def __init__(self):
        self._variables: Dict[str, Variable] = {}

    def __str__(self):
        return f"{self.__class__.__name__}:\n" + "\n".join(
            [f"{var.name}={var.value} {var.unit} ({var.description})"
             for var in self._variables.values()]
        )

    def set(self, name: str, value: float, unit: str = None, description: str = None):
        """
        Add or update a Variable in the container.

        Args:
            name (str): The name of the variable.
            value (float): The numerical value of the variable.
            unit (str): The unit of measurement for the variable (optional).
            description (str): A description of the variable (optional).
        """
        if name in self._variables:
            self._variables[name].value = value
        else:
            self._variables[name] = Variable(
                name=name, value=value, unit=unit, description=description
            )

    def __getattr__(self, item):
        """
        Overwrite the dunder method to enable usage of e.g.
        fs_state.COP
        """
        if item in {'__getstate__', '__setstate__'}:
            return super().__getattr__(self, item)
        if item in self._variables:
            return self._variables.get(item).value
        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{item}'")

    def get_variable_names(self) -> list:
        """
        Get the names of all variables in the container.

        Returns:
            list: A list of variable names.
        """
        return list(self._variables.keys())

    def get_variables(self):
        """
        Get all variables in the container.

        Returns:
            Dict[str, Variable]: A dictionary of variable names and Variable objects.
        """
        return self._variables

    def items(self):
        """
        Get items from the container.

        Returns:
            Dict[str, Variable]: A dictionary of variable names and Variable objects.
        """
        return self._variables.items()

    def get(self, name: str, default: Any = None):
        """
        Get the Variable with the specified name.

        Args:
            name (str): The name of the variable.
            default (Any): Default value to return if the variable is not found.

        Returns:
            Variable: The Variable object.

        """
        return self._variables.get(name, default)

    def copy(self):
        """
        Return a deepcopy of the instance as the variable dict is mutable.

        Returns:
            VariableContainer: A deepcopy of the VariableContainer instance.
        """
        return deepcopy(self)

    def convert_to_str_value_format(self, with_unit_and_description: bool) -> Dict[str, float]:
        """
        Returns a dictionary with a str: float format which is handy when storing results
        in files like .csv or .xlsx.

        Args:
            with_unit_and_description (bool): When False, only the name is in the string.

        Returns:
            dict: Containing all variables and values.
        """
        if with_unit_and_description:
            return {f"{k} in {v.unit} ({v.description})": v.value for k, v in self.items()}
        return {k: v.value for k, v in self.items()}


class FlowsheetState(VariableContainer):
    """
    This class is used to define the unique states of the flowsheet
    in the heat pump.

    The class is dynamic in the sense that attributes may be
    added during calculation of new flowsheet. This enables
    the easy adding of custom values to analyze the whole flowsheet
    and not restrict to a certain naming convention.
    """


class Inputs(VariableContainer):
    """
    Class defining inputs to calculate the FlowsheetState.

    While the inputs are pre-defined, you may add further ones
    using the `set` method.

    Args:
        n (float): Relative compressor speed between 0 and 1.
        T_eva_in (float): Secondary side evaporator inlet temperature.
        T_con_in (float): Secondary side condenser inlet temperature.
        m_flow_eva (float): Secondary side evaporator mass flow rate.
        m_flow_con (float): Secondary side condenser mass flow rate.
        dT_eva_superheating (float): Super-heating after evaporator.
        dT_con_subcooling (float): Subcooling after condenser.
        T_ambient (float): Ambient temperature of the machine.
    """

    def __init__(
            self,
            n: float = None,
            T_eva_in: float = None,
            T_con_in: float = None,
            m_flow_eva: float = None,
            m_flow_con: float = None,
            dT_eva_superheating: float = None,
            dT_con_subcooling: float = None,
            T_ambient: float = None
    ):
        """
        Initializes an Inputs object with parameters representing external conditions
        for the vapor compression cycle.

        Args:
            n (float): Relative compressor speed between 0 and 1 (unit: -).
            T_eva_in (float): Secondary side evaporator inlet temperature (unit: K).
            T_con_in (float): Secondary side condenser inlet temperature (unit: K).
            m_flow_eva (float): Secondary side evaporator mass flow rate (unit: kg/s).
            m_flow_con (float): Secondary side condenser mass flow rate (unit: kg/s).
            dT_eva_superheating (float): Super-heating after evaporator (unit: K).
            dT_con_subcooling (float): Subcooling after condenser (unit: K).
            T_ambient (float): Ambient temperature of the machine (unit: K).
        """
        super().__init__()
        self.set(
            name="n",
            value=n,
            unit="-",
            description="Relative compressor speed"
        )
        self.set(
            name="T_eva_in",
            value=T_eva_in,
            unit="K",
            description="Secondary side evaporator inlet temperature"
        )
        self.set(
            name="T_con_in",
            value=T_con_in,
            unit="K",
            description="Secondary side condenser inlet temperature"
        )
        self.set(
            name="m_flow_con",
            value=m_flow_con,
            unit="kg/s",
            description="Secondary side condenser mass flow rate"
        )
        self.set(
            name="m_flow_eva",
            value=m_flow_eva,
            unit="kg/s",
            description="Secondary side evaporator mass flow rate"
        )
        self.set(
            name="dT_eva_superheating",
            value=dT_eva_superheating,
            unit="K",
            description="Super-heating after evaporator"
        )
        self.set(
            name="dT_con_subcooling",
            value=dT_con_subcooling,
            unit="K",
            description="Subcooling after condenser"
        )
        if T_ambient is None:
            T_ambient = T_eva_in
        self.set(
            name="T_ambient",
            value=T_ambient,
            unit="K",
            description="Ambient temperature of machine"
        )



from .base import BaseCycle
from .standard import StandardCycle
from .vapor_injection_economizer import VaporInjectionEconomizer
from .vapor_injection_phase_separator import VaporInjectionPhaseSeparator



from abc import ABC
from vclibpy.media import ThermodynamicState, MedProp


class BaseComponent(ABC):
    """
    Abstract base class for defining interfaces of components in the vapor compression cycle.

    Methods:
        start_secondary_med_prop():
            To use multiprocessing, MedProp can't start in the main thread, as the object can't be pickled.
            This function starts possible secondary MedProp classes, for instance in heat exchangers.
            The default component does not have to override this function.

    Properties:
        state_inlet (ThermodynamicState):
            Property for accessing and setting the inlet state of the component.
        state_outlet (ThermodynamicState):
            Property for accessing and setting the outlet state of the component.
        m_flow (float):
            Property for accessing and setting the mass flow rate through the component.
        med_prop (MedProp):
            Property for accessing and setting the property wrapper for the working fluid.
    """

    def __init__(self):
        """
        Initialize the BaseComponent.
        """
        self._state_inlet: ThermodynamicState = None
        self._state_outlet: ThermodynamicState = None
        self._m_flow: float = None
        self._med_prop: MedProp = None

    def start_secondary_med_prop(self):
        """
        Start secondary MedProp classes for multiprocessing.

        To use multiprocessing, MedProp can't start in the main thread, as the object can't be pickled.
        This function starts possible secondary MedProp classes, for instance in heat exchangers.
        The default component does not have to override this function.
        """
        pass

    def terminate_secondary_med_prop(self):
        """
        To use multi-processing, MedProp can't start
        in the main thread, as the object can't be pickled.

        This function terminates possible secondary med-prop
        classes, for instance in heat exchangers.
        The default component does not have to override
        this function.
        """
        pass

    @property
    def state_inlet(self) -> ThermodynamicState:
        """
        Get or set the inlet state of the component.

        Returns:
            ThermodynamicState: Inlet state of the component.
        """
        return self._state_inlet

    @state_inlet.setter
    def state_inlet(self, state_inlet: ThermodynamicState):
        """
        Set the inlet state of the component.

        Args:
            state_inlet (ThermodynamicState): Inlet state to set.
        """
        self._state_inlet = state_inlet

    @property
    def state_outlet(self) -> ThermodynamicState:
        """
        Get or set the outlet state of the component.

        Returns:
            ThermodynamicState: Outlet state of the component.
        """
        return self._state_outlet

    @state_outlet.setter
    def state_outlet(self, state_outlet: ThermodynamicState):
        """
        Set the outlet state of the component.

        Args:
            state_outlet (ThermodynamicState): Outlet state to set.
        """
        self._state_outlet = state_outlet

    @property
    def m_flow(self) -> float:
        """
        Get or set the mass flow rate through the component.

        Returns:
            float: Mass flow rate through the component.
        """
        return self._m_flow

    @m_flow.setter
    def m_flow(self, m_flow: float):
        """
        Set the mass flow rate through the component.

        Args:
            m_flow (float): Mass flow rate to set.
        """
        self._m_flow = m_flow

    @property
    def med_prop(self) -> MedProp:
        """
        Get or set the property wrapper for the working fluid.

        Returns:
            MedProp: Property wrapper for the working fluid.
        """
        return self._med_prop

    @med_prop.setter
    def med_prop(self, med_prop: MedProp):
        """
        Set the property wrapper for the working fluid.

        Args:
            med_prop (MedProp): Property wrapper to set.
        """
        self._med_prop = med_prop



"""
Module with basic functions to calculate heat transfer coefficients.
"""
import abc

import numpy as np

from vclibpy.media import TransportProperties, ThermodynamicState, MedProp
from vclibpy.datamodels import FlowsheetState, Inputs


class HeatTransfer(abc.ABC):
    """
    Base class to implement possible heat transfer models.

    Methods:
        calc(transport_properties: TransportProperties, m_flow: float) -> float:
            Abstract method to calculate heat transfer.

    """

    @abc.abstractmethod
    def calc(self, transport_properties: TransportProperties, m_flow: float) -> float:
        """
        Calculate heat transfer.

        Args:
            transport_properties (TransportProperties): Transport properties of the medium.
            m_flow (float): Mass flow rate.

        Returns:
            float: Calculated heat transfer coefficient.

        Raises:
            NotImplementedError: If the method is not implemented in the subclass.
        """
        raise NotImplementedError


class TwoPhaseHeatTransfer(abc.ABC):
    """
    Base class to implement possible heat transfer models
    """

    @abc.abstractmethod
    def calc(
            self,
            state_q0: ThermodynamicState,
            state_q1: ThermodynamicState,
            state_inlet: ThermodynamicState,
            state_outlet: ThermodynamicState,
            med_prop: MedProp,
            inputs: Inputs,
            fs_state: FlowsheetState,
            m_flow: float
    ) -> float:
        """
        Calculate two-phase heat transfer.

        Args:
            state_q0 (ThermodynamicState): Thermodynamic state at the beginning of the two-phase region.
            state_q1 (ThermodynamicState): Thermodynamic state at the end of the two-phase region.
            state_inlet (ThermodynamicState): Inlet thermodynamic state.
            state_outlet (ThermodynamicState): Outlet thermodynamic state.
            med_prop (MedProp): Medium properties class.
            inputs (Inputs): Input parameters.
            fs_state (FlowsheetState): Flowsheet state.
            m_flow (float): Mass flow rate.

        Returns:
            float: Calculated two-phase heat transfer coefficient.

        Raises:
            NotImplementedError: If the method is not implemented in the subclass.
        """
        raise NotImplementedError


def calc_reynolds_pipe(dynamic_viscosity: float, m_flow: float, characteristic_length: float) -> float:
    """
    Calculate Reynolds number for flow inside a pipe.

    Args:
        dynamic_viscosity (float): Dynamic viscosity of the fluid.
        m_flow (float): Mass flow rate.
        characteristic_length (float): Characteristic length (e.g., diameter) of the pipe.

    Returns:
        float: Reynolds number.

    """
    return 4 * m_flow / (np.pi * characteristic_length * dynamic_viscosity)
